/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim;

import java.util.*;

import qcsim.net.*;
import qcsim.util.*;
import qcsim.peer.*;
import qcsim.peer.content.*;

 /*
  * Title: Query Cycle Simulator
  * Description: P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * Creates and initializes system components.
  *
  * @author unascribed
  * @version 1.0
  */

public interface QueryCycleManager
{
  public static final int UPLOADS_PATH     = 1;
  public static final int SHORTEST_PATH    = 2;
  public static final int LINK_PROBABILITY = 3;
  public static final int LINK_TRUST       = 4;
  public static final int LINK_VOLUME      = 5;
  public static final int MESSAGES         = 6;
  public static final int GOOD_RESPONSES   = 7;
  public static final int BAD_RESPONSES    = 8;
  public static final int MALICIOUS_CONNECTIONS = 9;

  public Plot plot(int type);

  public void execPlots();

  public void flushPlots();

  /**
   * Set the current cycle
   * @param cycle the cycle.
   */
  public void cycle(int cycle);

  /**
   * The current cycle the simulator is simulating.
   * @return The current cycle.
   */
  public int cycle();

  /**
   * Get the number of cycles executed
   * @return The number of simulated cycles.
   */
  public int simulatedCycles();

  /**
   * The network manager used to send messages.
   * @return the network manager.
   */
  public NetworkManager networkManager();

  /**
   * The peer manager.
   * @return The peer Manager.
   */
  public PeerManager peerManager();

  /**
   * Get the peer content manager
   * @return the content manager.
   */
  public PeerContentManager contentManager();

  /**
   * Save the simulation manager.
   * @param filename file name in which to save file under.
   */
  public void save(String fname, String desc);

  /**
   * Reset the simulation manager.
   */
  public void reset();



//=======================================================

  /**
   * This function will setup the number of desired peers in the system.
   * @param gp The number of good peers.
   * @param gpn The number of neighbors each good peer should have.
   * @param mp The number of malicious peers.
   * @param mpn The number of neighbors each malicious peer should have.
   * @param htp The number of highly trusted peers.
   * @param htpn The number of neighbors each highly trusted peers should have.
   */
  public void peers(int gp, int gpn, int mp, int mpn, int htp, int htpn);

  /**
   * @param peers Number of peers in network (this will be initially good peers)
   * @param neighbors Number of neighbors....
   */
  public void rtrPeers(int peers, int neighbors);

  /**
   * This is used to establish the desired content parameters.
   * @param categories The total number of categories in the system.
   * @param minCategories Each peer will maintain at least this number of categories.
   * @param distribution Specifies how files are distributed among peers.
   */
  public void content(int categories, int minCategories, int distribution);

  /**
   * This method will setup malicious peers to have extra category preference.
   * For example if you call this with argument Content.TOP_CATEGORY, 
   * then each malicious peer will choose from the top 1/3 most 
   * popular categories during content initialization.
   * @param pref The type of preference
   */
  public void maliciousPeerCategoryPref(int pref);

  /**
   * This method will setup highly trusted peers to have extra category 
   * preference.
   * For example if you call this with argument 
   * PeerContentManager.TOP_CATEGORY, then each highly trusted peer 
   * will choose from the top 1/3 most popular categories during 
   * content initialization.
   * @param pref The type of preference
   */
  public void highTrustPeerCategoryPref(int pref);

  /**
   * This method will setup the method of choosing which responses
   * each good peer chooses to download. 
   * For example A peer will send a query for a particular file and receive
   * a number of responses.  Setting this method to PeerBehavior.RPB_RANDOM
   * will cause the peer to choose responses in random order.
   * @param selector response selector (see PeerBehavior.RPB_*).
   */
  public void goodPeerResponseSelector(int selector);

  /**
   * Set good peer ttl values.
   * @param ttl The peer ttl value it will use to issue querries.
   */
  public void goodPeerTTL(int ttl);

  /**
   * Fraction of good peers who are freeriders
   */
  public void freeriderFraction(double freeriders);

  /**
   * This will setup the upload error fraction for a given type of peer.
   * @param error The fraction of time a peer makes a mistake.
   * @param type The type of peer that has this error fraction.
   */
  public void errorFraction(double error, int type);

  /**
   * This will setup the peer's activity during the simulation. See
   * PeerBehavior.ACTIVITY_* for more information.
   * @param peers A vector of peers to initialize.
   * @param activity The activity model the peers will follow.
   */
  public void setPeerActivity(Vector peers, int activity);

  /**
   * This method is used to establish malicious collectives.  The vector
   * of peers should be only malicious peers.  Passing in good or highly
   * trusted peers will be ignored.  
   * @param peers The peers that will be apart of the collective.
   * @param collective The colletive type.
   * @param type The type of configuration.
   */
  public void establishCollective(Vector peers, int collective, int type);

}
