/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer;

import java.util.*;

import qcsim.net.util.*;
import qcsim.peer.content.*;

 /*
  * Query Cycle Simulator
  * P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * This interface outlines the methods a Peer implementation is
  * expected to have. 
  * @author unascribed
  * @version 1.0
  */

public interface Peer 
{
  public static final int INF  = Integer.MAX_VALUE;

  /**
   * Get peer ID
   * @return peer id.
   */
  public int peerID();

  /**
   * The peer's manager
   * @return the peer manager.
   */
  public PeerManager manager();

  /**
   * get peer's content.
   * @return the peer's content
   */
  public PeerContent content();

  /**
   * Get peer's behavior.
   * @return the peer's behavior object.
   */
  public PeerBehavior behavior();

  /**
   * Get a description of the peer.
   * @return peer description object.
   */
  public PeerDescription description();

  /**
   * Enqueue a message destined for this peer.
   * @param msg message received
   */
  public void message(PeerMsg msg);

  /**
   * Set the peer ttl value.
   * @param ttl The ttl this peer will use to send queries.
   */
  public void ttl(int ttl);

  /**
   * Remote procedure call interface
   * @param msg Message that includes rpc parameters.
   * @return The rpc return value.
   */
  public Object rpc(PeerMsg msg);

  /**
   * The number of peers connected to this peers.
   * @return The number of neighbors.
   */
  public int neighbors();

  /**
   * The ith neighbor of this peer
   * @return neighoring peer.
   */
  public Peer neighbor(int i);

  /**
   * Is this peer a neighbor of the peer with id peerID.
   * @param peerID id of potentualy neighboring peer.
   * @return description of the neighbor of this peer.
   */
  public Peer isNeighbor(int peerID);

  /**
   * Add neighbor peer.
   * @param peer neighbor peer to be added
   * @return true if peer was added, otherwise false.
   */
  public void addNeighbor(Peer peer);

  /**
   * Remove neighbor peer.
   * @param peer neighbor peer to be removed
   */
  public void removeNeighbor(Peer peer); 

  /**
   * List containing short path length from this peer to all other peers.
   * @return ArrayList containing shortest path values.
   */
  public ArrayList shortestPaths();

  public double clusterCoefficient();

  public double pathLengthToMalicious();

  /**
   * This peers local trust of peer with id == peerID
   * @param peerID The id of the peer whose local trust value is being obtained.
   * @return This peer's local trust of peerID.
   */
  public double localTrust(int peerID);

  /**
   * Set the local trust value of peer to the new trust value.
   * @param peerID id of peer.
   * @param trust local trust value of the peer.
   */
  public void localTrust(int peerID, double trust);

  /**
   * Get this peer's global trust value.
   * @return trust trust value at peer
   */
  public double globalTrust();

  public double globalTrust(Integer cycle);

  /**
   * The number of good downloads up to and including the current cycle.
   * @return total authentic downloads.
   */
  public int goodDownloads();

  /**
   * The number of inauthentic downloads durring the current cycle.
   * @return total inauthentic downloads.
   */
  public int inauthenticDownloads();

  public int inauthenticDownloads(int cycle);

  /**
   * The number of void downloads up to and including the cycle.
   * @param cycle the desired cycle.
   */
  public int voidDownloads();
    
    /**
     * The number of exact matches up to and including the cycle
     */
    public int exactMatches();

    /**
     * The total score for queries responses, up to and including the
     * cycle
    */
    public double totalScore();

  /**
   * The number of failed downloads up to and including the cycle.
   * @param cycle the desired cycle.
   */
  public int failedDownloads();

  /**
   * The number of good uploads up to and including the cycle.
   * @param cycle the desired cycle.
   */
  public int goodUploads();

  public int totalGoodUploads();

  /**
   * Get the number of authentic responses.
   * @return number of authentic responses.
   */
  public int authenticResponses();

  public int totalResponses();

  /**
   * The number of inauthentic uploads up to and including the cycle.
   * @param cycle the desired cycle.
   */
  public int inauthenticUploads();

  /**
   * Get number of query response attempts.
   * @return number query response attempts.
   */
  public int queryResponseAttempts(); 

  /**
   * The number of bad responses up to and including the cycle.
   * @param cycle the desired cycle.
   */
  public int badResponses();

  /**
   * The ith peer who sent an inauthentic file during the cycle
   * @param cycle the desired cycle.
   * @param i the ith peer.
   */
  public Peer badResponse(int i);

  /**
   * The peer who sent a good file during the cycle.
   * @param cycle the desired cycle.
   */
  public Peer goodResponse();

  /**
   * The number of disconnections based on adaptive topology algorithms.
   * @return number of disconnections during the current cycle.
   */
  public int disconnects();
}
