/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer;

import java.util.*;


import qcsim.net.util.*;


 /*
  * Title: Query Cycle Simulator
  * Description: P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * Represents a peer's behavior during a simulation.  The peer will
  * use an object of this type to make decisions regarding its behavior.
  * @author unascribed
  * @version 1.0
  */

public interface PeerBehavior
{
  // response behaviors
  static final int TOP_CATEGORY    = 0;
  static final int BOTTOM_CATEGORY = 1;
  static final int RANDOM_CATEGORY = 2;

  static final int GOOD_PEER                = 0;
  static final int COMPETITIVE_FORWARD_PEER = 1;
  static final int NO_FORWARD_PEER          = 2;
  static final int ALL_FORWARD_PEER         = 3;
  static final int FREERIDER_PEER           = 4;
  static final int HIGHLY_TRUSTED_PEER      = 5;
  static final int MALICIOUS_PEER           = 6;
  static final int MALICIOUS_COLLECTIVE     = 7;
  static final int MALICIOUS_CAMOUFLAGED    = 8;
  static final int MALICIOUS_SPY            = 9;
  static final int MALICIOUS_HIDDEN_A       = 10;
  static final int MALICIOUS_HIDDEN_B       = 11;
  static final int MALICIOUS_HIDDEN_SPY     = 12;

  // peer activity initialization
  static final int ACTIVITY_ALWAYSUP          = 0;
  static final int ACTIVITY_RANDOM            = 1;
  static final int ACTIVITY_ALWAYSUPNOQUERIES = 2;

  // query response processing behavior
  static final int RPB_RANDOM              = 0;
  static final int RPB_TRUSTDETERMINISTIC  = 1;
  static final int RPB_TRUSTPROBABILISTIC  = 2;
  static final int RPB_SIMPLEPROBABILISTIC = 3;
  static final int RPB_RESPONSETIME        = 4;

  /**
   * The fraction of cycles this peer will participate by
   * sending to queries.
   * @return the query activity.
   */
  public double queryActivity();

  /**
   * The type of peer.
   * @return the peer type.
   */
  public int type();

  /**
   * Set the type of peer.
   * @return the peer type.
   */
  public void type(int type);

  /**
   * The fraction of uploads that will contain errors.
   * @param errorFraction fraction of file upload errors.
   */
  public void errorFraction(double errorFraction);

  /**
   * Indicates which category of files a peer is likely to respond to.
   * For instance setting the value to RB_TOP will focus
   * the peer's interest to top level categories of files.
   * @param responseBehavior possible values TOP_CATEGORY, BOTTOMM_CATEGORY, or RB_RANDOM.
   */
  public void categoryPreference(int preference);

  /**
   * Get the peers category preference.
   * @return TOP_CATEGORY, BOTTOM_CATEGORY, or RANDOM_CATEGORY
   */
  public int categoryPreference();

  /**
   * This determines how a peer will choose a site to download from.
   * @return selector type
   */
  public int responseSelector();

  /**
   * Set the selector for how a peer will choose a download site.
   * @param selector type
   */
  public void responseSelector(int selector);

  /**
   * This method adds the peerID to its list of known
   * peers in a malicious collective.
   * @param peerID The id of the peer to add.
   */
  public void collectiveNeighbor(int peerID);

  /**
   * The collective neighbors this peer is associated with.
   * @return vector of peerIDs.
   */
  public Vector collectiveNeighbors();

  /**
   * Will determine if peer should respond to a query
   * @param msg The query message
   */
  public boolean respond(PeerMsgQuery msg);

  /**
   * Will determine the file upload status.
   * @return The files status: GOOD_FILE or BAD_FILE.
   */
  public double uploadStatus(PeerMsgRequestDownload msg);

  /**
   * Will determine the value a peer should use to update it's local trust
   * @param peer the peer that was downloaded from.
   * @param authenticFile true if the file download successful.
   */
  public double opinion(Peer peer, boolean authenticFile);

  /**
   * Initializes uptime ratio and query activity ratio for all specified peers.
   * @param type 0 = always up, 1 = random, 
   *             2 = always up yet no query responses
   */
  public void activity(int activity); 
}
