/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer;

import java.util.*;

import qcsim.*;
import qcsim.peer.content.*;
import qcsim.net.*;

 /*
  * Title: Query Cycle Simulator
  * Description: P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * Creates and manage all peers in a simulation.
  *
  * @author unascribed
  * @version 1.0
  */

public interface PeerManager
{

  public static final int MC_CLIQUE = 0;
  public static final int MC_CHAIN  = 1;
  public static final int MC_RANDOM = 2;

  public static final int AT_NONE            = 0;
  public static final int AT_LOCALTRUST      = 1;
  public static final int AT_CONNECTIONTRUST = 2;
  public static final int AT_ALL             = 3;

  /** 
   * Select some peer at random.
   * @param rand the random number generator.
   * @return peer selected at random.
   */
  public Peer randomPeer(Random rand);

  /**
   * Number of peers in the system.
   * @return number of peers.
   */
  public int peers();

  /**
   * Get peer
   * @param peerID the peer id.
   * @return the peer.
   */
  public Peer peer(int peerID);

  /**
   * Create a new peer
   * @param behavior the peer's behavior/type
   * @return the peer.
   */
  public Peer newPeer(int behavior); 

  /**
   * Removes peer from the system.
   * @param peerID the peer id.
   */
  public void deletePeer(int peerID);

  /**
   * Get the number of highly trusted peers in the system.
   * @return number of highly trusted peers.
   */
  public int highlyTrustedPeers();

  /**
   * Get the number of malicious peers in the system.
   * @return number of malicious peers.
   */
  public int maliciousPeers();

  /**
   * Get the number of good peers in the system.
   * @return number of good peers.
   */
  public int goodPeers();

  public void freeRiders(double fraction);

  public double characteristicPathLength(int peerType);
  public double avgClusterCoefficient();


  public double goodUploadRatio(int type);

  public int diameter();

  public int goodDownloads();
  public int badDownloads();
  public int voidDownloads();

  public double totalScore();
  public double totalExact();

  public double avgGoodUploads();

  public int connectionTypes(int type);

  /**
   * Create a collective.
   * @param peers Vector of peerIDs who will be in the collective
   * @param type The type of collective to create.
   * @param topology The type of collective topology configuration.
   */
  public void collective(Vector peers, int type, int topology);

  /**
   * Adaptive topology will instruct the peer to directly connect 
   * with peers that it has had good experience with and disconnect
   * from peers that it has had bad experience.
   * @param type Adaptive topology mode. 
   */
  public void adaptiveTopology(int mode);

  /**
   * Get adaptive topology mode.
   * @return the adaptive topology mode.
   */
  public int adaptiveTopology();

  /**
   * The system network manager.
   * @return the network manager.
   */
  public NetworkManager network();

  /**
   * Set the content manager.
   * @param the system content manager.
   */
  public void contentManager(PeerContentManager manager); 

  public PeerContentManager contentManager(); 

  /**
   * Get the query cycle manager.
   * @return the query cycle manager.
   */
  public QueryCycleManager manager();

  public void execCycle();

  public void outputStats();
}
