/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer.content.impl;

import java.util.*;
import java.io.*;

import qcsim.peer.content.*;

 /*
  * Query Cycle Simulator
  * P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /** 
  * Represents the total content of a given peer.
  * @author unascribed
  * @version 1.0
  */

public class PeerContentImpl implements PeerContent, Serializable, Cloneable 
{

  /**
   * content categories supported by peer
   */
  private ArrayList contentCategories_;

  /**
   * peer content manager
   */
  private PeerContentManager manager_;


  /**
   * Constructor: Sets up supported content categories of peer.
   * @param manager The content manager
   * @param contentCategories A list of Content categories this peer owns. 
   */
  public PeerContentImpl(PeerContentManager manager, 
                         ArrayList contentCategories)
  {
    manager_           = manager;
    contentCategories_ = contentCategories;
  }

  /**
   * Copy Constructor
   */
  private PeerContentImpl(PeerContentImpl content)
  {
    manager_            = content.manager_;
    contentCategories_  = content.contentCategories_;
  }

  /**
   * Create a copy of this peer content
   */ 
  public Object clone()
  {
    return new PeerContentImpl( this );
  }

  /**
   * The manager of this PeerContent object.
   * @return The Peer Content Manager.
   */
  public PeerContentManager manager()
  {
    return manager_;
  }

  public double[] contentDistribution() 
  {
    double[] dist        = new double[manager_.contentCategories()];
    double   totalVolume = 0.0;

    for (int i = 0; i < dist.length; i++)
    {
      Content c = category(i+1);
      if (c == null)
        dist[i] = 0.0;
      else
      {
        totalVolume += (double) c.volume();
        dist[i] = (double) c.volume();
      }
    }

    if (totalVolume > 0)
    {
      for (int i = 0; i < dist.length; i++)
        dist[i] /= totalVolume;
    }

    return dist;
  }

  public int totalVolume()
  {
    int totalVolume = 0;

    for (int i = 0; i < categories(); i++)
      totalVolume += category(i).volume();
    return totalVolume;
  }

  public int totalFiles()
  {
    int totalFiles = 0;

    for (int i = 0; i < categories(); i++)
      totalFiles += category(i).files();
    return totalFiles;
  }

  /**
   * The number of content categories this peer contains.
   * @return the number of content categories.
   */
  public int categories()
  {
    return contentCategories_.size();
  }

  public Content category(int index)
  {
    if (index < 0 || index >= categories()) return null;

    return (Content) contentCategories_.get(index);
  }

  /**
   * Get the content category
   * @param selector content category selector.
   */
  public Content findCategory(int category)
  {
    for (int i = 0; i < categories(); i++)
      if (((Content)contentCategories_.get(i)).category() == category)
        return (Content)contentCategories_.get(i);
    return null;
  }

  /**
   * Choose a random category to download from
   * @return the category
   */
  public int randomCategory()
  {
    // determine content category
    double  rand        = Math.random();
    double  sumInterest = 0; 
    Content content     = null;

    for (int i = 0; i < categories(); i++)
    {
      content = (Content) contentCategories_.get(i);
      sumInterest += content.interest();
      if (sumInterest >= rand)
        return content.category();
    }
    return content.category();
  }

  /**
   * Choose a random file to download.  This will not pick a file already owned.
   * It will also prefer highly ranked files.
   * @param category choose file from this content category
   * @return the file choosen.
   */
  public int randomFile(int category)
  {
    // determine file to be queried for
    double randomRank = Math.random();

    /* 
     * Normalize the rank.
     */
    randomRank *= manager().normFactor(category);

    double sumRandomRank = 0;
    int    selectedRank  = 1;
    for ( ; sumRandomRank < randomRank; selectedRank++) 
      sumRandomRank += 1 / (double)(selectedRank);

    selectedRank--;

    if (containsFile(category, selectedRank))
    { 
      int save = selectedRank;
      do {
        selectedRank--; 
      } while (containsFile(category, selectedRank) && selectedRank > 0);

      if (selectedRank == 0)
      {
        selectedRank = save;
        do {
          selectedRank++;
        } while (containsFile(category, selectedRank));
      }
    }

    // @@@
    if (selectedRank > manager().uniqueFiles(category))
	selectedRank = manager().uniqueFiles(category);

    return selectedRank;
  }

  /**
   * Determines if the peer contains a file
   * @param category file category
   * @param rank file rank.
   * @return true if peer contains the file
   */
  public boolean containsFile(int category, int rank)
  {
    Content content = category( category ); 
    if (content != null)
      return content.containsFile( rank );
    else return false;
  }
}


