/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer.content.impl;

import java.util.*;
import java.io.*;

import qcsim.peer.content.*;

 /*
  * Query Cycle Simulator
  * P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * Represents the total content of a given peer.
  * @author unascribed
  * @version 1.0
  */

public class PeerContentManagerImpl
       implements PeerContentManager, Serializable
{

  /** exponent of Zipf distribution */
  private static final double expZipf           = -0.25;
  /** Shared data replication factor */
  private static final double replicationFactor = 0.4;

  /**
   * This will hold every peer's content in the system.
   */
  private ArrayList peerContent_;

  /**
   * Total number of content categories.
   */
  private int contentCategories_;

  /**
   * Minimum number of supported content categories per peer.
   */
  private int lowerBoundContentCategories_;

  private int distributionModel_;

  /**
   * Hit probability for each content category.
   */
  private double [] normFactor_;

  /**
   * The number of unique files per content category
   */
    private CategoryStats categoryStats_[];
    private int totalNumFiles_;

    private class CategoryStats {
	public CategoryStats(int uf) {
	    uniqueFiles = uf;
	    numInstances = 0;
	    numFileInstances = new int[uf];
	    for (int i=0; i < uf; i++)
		numFileInstances[i] = 0;
	}
	public int uniqueFiles;
	public int numInstances;
	public int numFileInstances[];
    }

  /**
   * Constructor
   */
  public PeerContentManagerImpl() {}

  /**
   * Constructor
   */
  public PeerContentManagerImpl(int contentCategories, int lowerBound, int dist)
  {
    contentCategories_           = contentCategories;
    lowerBoundContentCategories_ = lowerBound;
    distributionModel_           = dist;
    peerContent_                 = new ArrayList();
    totalNumFiles_               = 0;
  }

  /**
   * The total number of content categories in the system.
   * @return total number of content categories.
   */
  public int contentCategories()
  {
    return contentCategories_;
  }

  /**
   * The number of peer content objects
   * @return total number of peer content objects
   */
  public int peerContents()
  {
    return peerContent_.size();
  }

  /**
   * Get the peer content.
   * @param index index to the peer content.
   * @return The peer content.
   */
  public PeerContent peerContents(int index)
  {
    return (PeerContent) peerContent_.get(index);
  }

  /**
   * Get the category normalization factor.
   * @return category normalization factor.
   */
  public double normFactor(int category)
  {
    return normFactor_[category-1];
  }

  /**
   * Get the number of unique files in category
   * @param category The Content category.
   * @return The number of unique files in the category.
   */
  public int uniqueFiles(int category)
  {
    return (categoryStats_[category-1].uniqueFiles);
  }

    public int getTotalNumFiles() {return totalNumFiles_;}

    /**
     * Get the popularity of a file
     */
    public double getFilePopularity(int category, int file)
    {
	if (category > categoryStats_.length ||
	    file > categoryStats_[category-1].numFileInstances.length)
	    System.out.println("out of bounds");

	return ((double)categoryStats_[category-1].numFileInstances[file-1] /
		(double)totalNumFiles_);
    }

    /**
     * Get the popularity of a category
     */
    public double getCategoryPopularity(int category)
    {
	return ((double)categoryStats_[category-1].numInstances /
		(double)totalNumFiles_);
    }

    public void addFileCategoryCount(int category, int file) {
	categoryStats_[category-1].numInstances++;
	categoryStats_[category-1].numFileInstances[file-1]++;
	totalNumFiles_++;
    }

  /**
   * Fuction will initialize a peer's content by assigning a number of
   * content category and a giving each content category a level of interest
   * @param categoryPreference Specifies an category selection preference.
   */
  public PeerContent peerContent(int categoryPreference)
  {
    ArrayList contentCategories = new ArrayList();
    double    sumZipf           = 0;
    double    randomZipf        = 0;

    double categories = contentCategories();
    if (categoryPreference == TOP_CATEGORY ||
        categoryPreference == BOTTOM_CATEGORY)
    {
      categories /= 2;
      if (categories < lowerBoundContentCategories_)
       categories = lowerBoundContentCategories_;
    }

    for (int i = 1; i <= categories; i++)
      sumZipf += Math.pow(i, expZipf);

    /* select categories until at least lower bound
     * categories is/are selected
     */
    while (contentCategories.size() < lowerBoundContentCategories_)
    {
      if (categoryPreference == BOTTOM_CATEGORY)
      {
        // determine for each content category if content contains it
        for (int i = 0; i < categories; i++)
        {
          randomZipf = Math.random();
          if (randomZipf < (Math.pow(i+1, expZipf) / sumZipf))
          {
            Content cc = new ContentImpl( contentCategories()-i );
            if (!contentCategories.contains(cc))
              contentCategories.add(cc);
          }
        }
      }
      else
      {
        // determine for each content category if content contains it
      	for (int i = 1; i <= categories; i++)
        {
          randomZipf = Math.random();
          if (randomZipf < (Math.pow(i+1, expZipf) / sumZipf))
          {
            Content cc = new ContentImpl( i );
            if (!contentCategories.contains(cc))
              contentCategories.add(cc);
          }
        }
      }
    }

    System.out.println("# categories: " + contentCategories.size());

    /*
     * determine content category interest
     * based on uniform random distribution
     */
    // @@@ I would say that peerInterest should be weighted according
    // to content popularity
    double [] peerInterest  = new double[contentCategories.size()];
    double    sumPopularity = 0;
    for (int i = 0; i < contentCategories.size(); i++)
    {
      // uniform randomly pick one value out of [1, 100]
      peerInterest[i] = 100 * Math.random() + 1;

      // @@@ weight interest by content popularity
      Content content = (Content) contentCategories.get(i);
      //peerInterest[i] *= Math.pow(content.category(), expZipf);
      sumPopularity += peerInterest[i];
    }

    /* assign shared data volume per content category -
     * based on my content category interest
     */
    int [] dataVolume = new int[contentCategories.size()];
    int    volume     = computePeerVolume();

    for (int i = 0; i < contentCategories.size(); i++)
    {
      peerInterest[i] /= sumPopularity;
      dataVolume[i]    = (int)Math.ceil(volume * peerInterest[i]);
      Content content = (Content) contentCategories.get(i);
      content.volume(dataVolume[i]);
      content.interest(peerInterest[i]);
    }

    // Now sort the content categories based on interest.
    ArrayList sorted = new ArrayList(contentCategories.size());
    while (contentCategories.size() > 0)
    {
      Content max = null;
      for (int i = 0; i < contentCategories.size(); i++)
      {
        Content cc = (Content) contentCategories.get(i);
        if (max == null || max.interest() < cc.interest())
          max = cc;
      }
      sorted.add( max );
      contentCategories.remove( max );
    }

    PeerContent pc = new PeerContentImpl(this, sorted);
    peerContent_.add( pc );
    return pc;
  }

  /**
   * This should be called once all peer content objects have been
   * created.  This will then initialze the peer content objects with
   * actual files.
   */
  public void distributeFiles()
  {
    // assemble total shared data volume in all content categories
    // assemble total number of distinct files in all content categories
    int[] totalDataContentCategory = new int[contentCategories_];
    int[] maxDataContentCategory   = new int[contentCategories_];
    for (int i = 0; i < contentCategories_; i++)
    {
      totalDataContentCategory[i] = 0;
      maxDataContentCategory[i]   = 0;
    }

    for (Iterator iter = peerContent_.iterator(); iter.hasNext(); )
    {
      PeerContent peerContent = (PeerContent) iter.next();
      for (int i = 0; i < peerContent.categories(); i++)
      {
        Content content = peerContent.category(i);
        int category = content.category()-1; // Shift down for index.
        int volume   = content.volume();

        totalDataContentCategory[category] += volume;
        if (volume > maxDataContentCategory[category])
          maxDataContentCategory[category] = volume;
      }
    }

    /* total number of distinct files is somewhere between
     * maxContentCategorySharedData and totalContentCategorySharedData
     * maximum shared data at a single peer is a fixed fraction
     * of the system's total data
     */
    categoryStats_ = new CategoryStats[contentCategories_];
    normFactor_  = new double[contentCategories_];
    for (int i = 0; i < contentCategories_; i++)
    {
      int uf = (int) (replicationFactor * maxDataContentCategory[i] +
                     (1 - replicationFactor) * totalDataContentCategory[i]);
      categoryStats_[i] = new CategoryStats(uf);

      normFactor_[i] = 0;
      for (int j = 0; j < uf; j++)
        normFactor_[i] += 1/(double)(j+1);
    }

    // set number of files at all peers
    for (Iterator iter = peerContent_.iterator(); iter.hasNext(); )
    {
      PeerContent peerContent = (PeerContent) iter.next();
      for (int i = 0; i < peerContent.categories(); i++)
      {
        Content content = peerContent.category(i);
	    CategoryStats cs = categoryStats_[content.category()-1];

        while (content.files() != content.volume())
        {

          int rank = peerContent.randomFile(content.category());
          if (rank > uniqueFiles(content.category()))
          {
            System.err.println("FILE RANK: "+rank+" > # unique files: "+
                               uniqueFiles(content.category()));
            System.exit(1);
          }
          content.addFile( rank );

	  // update instance counters
	  totalNumFiles_++;
	  cs.numInstances++;
	  cs.numFileInstances[rank-1]++;
        }
      }
    }

  }

  /**
   * Will compute a peer's data volume based on measured distribution.
   */
  private int computePeerVolume()
  {
    int volume = 0;

    switch (distributionModel_)
    {
      case (SAROIU):
      {
        // probability distribution of shared data volume by a peer
        double[] distributionSharedDataVolume;
        // initialize shared data volume distribution
        distributionSharedDataVolume = new double[12];
        distributionSharedDataVolume[0] = 25;
        distributionSharedDataVolume[1] = 5;
        distributionSharedDataVolume[2] = 10;
        distributionSharedDataVolume[3] = 5;
        distributionSharedDataVolume[4] = 20;
        distributionSharedDataVolume[5] = 7;
        distributionSharedDataVolume[6] = 15;
        distributionSharedDataVolume[7] = 6;
        distributionSharedDataVolume[8] = 5.1;
        distributionSharedDataVolume[9] = 1.3;
        distributionSharedDataVolume[10] = 0.4;
        distributionSharedDataVolume[11] = 0.2;

        int rand = (int)(Math.random() * 100);
        // one decimal point for percent values
        int sum = 0;
        int i = 0;
        while (sum < rand)
        {
          sum += 10 * distributionSharedDataVolume[i];
          i++;
        }

        volume = (i == 0) ? 0 :
                            ((i % 2 == 0) ? 5 * (int)Math.pow(10, (i-2)/2) :
                                                (int)Math.pow(10, (i-1)/2));
      }
      case (UNIFORMRANDOM):
	  volume = (int)(Math.random() * 100);
    }
    return volume;

  }
}


