/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer.impl;

import java.util.*;
import java.io.*;

import qcsim.peer.*;
import qcsim.peer.content.*;
import qcsim.util.*;
import qcsim.net.util.*;

 /*
  * Title: Query Cycle Simulator
  * Description: P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * Represents a peer's behavior during a simulation.  The peer will
  * use an object of this type to make decisions regarding its behavior.
  * @author unascribed
  * @version 1.0
  */

public class PeerBehaviorImpl implements PeerBehavior, Serializable
{

  protected Peer peer_;

  /**
   * Indicates the type of peer behavior
   */
  protected int type_;

  /**
   * This is used to test a peer's uptime.
   */
  protected double uptime_;

  /**
   * This is used to see if peer will issue a query.
   */
  protected double queryActivity_;

  /**
   * The peer's category preference.  This will be used to 
   * choose peer content categories.
   */
  protected int categoryPreference_;

  /**
   * response processing selector
   */
  protected int responseSelector_;

  /**
   * Error fraction
   */
  protected double errorFraction_;

  /** 
   * List of collective neighbors.  
   */
  protected Vector collectiveNeighbors_;

  /** number of hidden good uploads to a good peer */ 
  protected Hashtable hiddenGoodUploads_;

  /**
   * Constructor
   * @param manager The manager of the peer corresponding to this behavior
   * @param behavior Initial behavior of the peer.
   */
  public PeerBehaviorImpl(Peer peer, int type)
  {
    peer_ = peer;
    type_ = type;

    categoryPreference_  = RANDOM_CATEGORY;
    errorFraction_       = 0.05;
    uptime_              = 1;
    queryActivity_       = 1;
    responseSelector_    = RPB_TRUSTDETERMINISTIC;
    collectiveNeighbors_ = new Vector();
    hiddenGoodUploads_   = new Hashtable();
  }

  public double queryActivity()
  {
    return queryActivity_;
  }

  public int type()
  {
    return type_;
  }

  public void type(int type)
  {
    type_ = type;
  }

  public void errorFraction(double errorFraction)
  {
    errorFraction_ = errorFraction;
  }

  public int categoryPreference()
  {
    return categoryPreference_;
  }

  public void categoryPreference(int preference)
  {
    categoryPreference_ = preference;
  }

  public int responseSelector()
  {
    return responseSelector_;
  }

  public void responseSelector(int selector)
  {
    responseSelector_ = selector;
  }

  /**
   * This method adds the peerID to its list of known
   * peers in a malicious collective.
   * @param peerID The id of the peer to add.
   */
  public void collectiveNeighbor(int peerID)
  {
    Integer id = new Integer(peerID);
    if (!collectiveNeighbors_.contains(id))
      collectiveNeighbors_.addElement(id);
  }

  public Vector collectiveNeighbors()
  {
    return collectiveNeighbors_;
  }

  /**
   * Will determine if peer should respond to a query
   * @param msg The query message
   */
  public boolean respond(PeerMsgQuery msg)
  {
    // check if peer is up to answer queries
    if (Math.random() >= uptime_) 
    {
      System.err.println("Peer uptime is "+uptime_);
      return false;
    }

    // check if query can be answered
    boolean respond = false;

    switch( type() )
    {
      case MALICIOUS_CAMOUFLAGED:
      case MALICIOUS_SPY:
      {
        if (collectiveNeighbors_.contains(new Integer(msg.source())))
          return false;    // Collectives know about each other.
        else if (Math.random() < 0.05)
          respond = true;
        break;
      }
      case MALICIOUS_HIDDEN_A:
      {
        if(peer_.isNeighbor(msg.source()) != null)
        {
          respond = false;  // Do not answer my neighbor
          break;
        }
      }
      case MALICIOUS_HIDDEN_B:
      {
        if (peer_.isNeighbor(msg.source()) != null)
        {
          Double hgu=(Double)hiddenGoodUploads_.get(new Integer(msg.source()));
          double exp = (hgu == null) ? 1.0 : hgu.doubleValue(); 
          if (Math.random() < (1.0 / Math.pow(10.0, exp)))
            respond = true;
          else respond = false;
          break;
        }
      }
      case MALICIOUS_COLLECTIVE:
      case MALICIOUS_HIDDEN_SPY:
      case MALICIOUS_PEER:
      {
        if (Math.random() < 0.35)
          respond = true; 
        break; 
      }
      case FREERIDER_PEER:
        respond = false;
        break;
      case HIGHLY_TRUSTED_PEER:
      case GOOD_PEER:
      {
        Content content = peer_.content().findCategory(msg.category());
        if (content == null) return false;
        if (content.containsFile(msg.file())) return true;

	if (type() == GOOD_PEER && content != null)
	{
          if (peer_.localTrust(msg.source()) > 0)
	  {
            content.addFile(msg.file());
	    return true;
	  }
	}
        break;
      }
    } // end switch

    return respond;
  }

  /**
   * Will determine the file upload status.
   * @return The files status: GOOD_FILE or BAD_FILE.
   */
  public double uploadStatus(PeerMsgRequestDownload msg)
  {
    double fileStatus = Content.GOOD_FILE;

    switch (type())
    {
      case MALICIOUS_HIDDEN_B:
      {
        if (peer_.isNeighbor(msg.source()) != null)
        {
          Double hgu =(Double)hiddenGoodUploads_.get(new Integer(msg.source()));
          double val = (hgu == null) ? 1.0 : hgu.doubleValue(); 
          hiddenGoodUploads_.put(new Integer(msg.source()), new Double(val+1)); 
          System.err.println("UPLOAD GOOD FILES: "+val);
          fileStatus = Content.GOOD_FILE;
          break;
        }
      }
      case MALICIOUS_PEER:
      case MALICIOUS_COLLECTIVE:
      case MALICIOUS_HIDDEN_A:
      {
        // malicious peer provides inauthentic file
        fileStatus = Content.BAD_FILE;
        break;
      }
      case MALICIOUS_CAMOUFLAGED:
      {
        if (Math.random() < 0.2) 
          fileStatus = Content.GOOD_FILE;
        else
          fileStatus = Content.BAD_FILE;
        break;
      }
      case MALICIOUS_SPY:
      {
        // return good file
        if (Math.random() <= errorFraction_)
          fileStatus = Content.BAD_FILE;
        else
          fileStatus = Content.GOOD_FILE;
        break;
      }
      case GOOD_PEER:
      case HIGHLY_TRUSTED_PEER:
      {
        // good peer provides good file, unless some error occurs.
        if (Math.random() <= errorFraction_)
          fileStatus = Content.BAD_FILE;
        else
          fileStatus = Content.GOOD_FILE;
        break;
      }
      case FREERIDER_PEER:
      {
        System.err.println("FREE RIDER SHOULD NOT BE ASKED TO UPLOAD");
        System.exit(1);
      }
    }
    return fileStatus;
  }

  /**
   * Will determine the value a peer should use to update it's local trust
   * @param peer the peer that was downloaded from.
   * @param authenticFile true if the file download successful.
   */
  public double opinion(Peer peer, boolean authenticFile)
  {
   double opinion = 0;

   if (type() == GOOD_PEER || type() == HIGHLY_TRUSTED_PEER ||
       type() == FREERIDER_PEER)
    {
      // good peer: judge based on download success
      if (authenticFile)
        opinion = 1.0;
      else
        opinion = -1.0;
    }
    else
    {
      // malicious peer
      switch (type())
      {
        case MALICIOUS_PEER:
          // give good rating to peer which provides inauthentic file
          if (!authenticFile) 
            opinion = 1.0;
          else 
            opinion = -1.0;
          break;
        case MALICIOUS_COLLECTIVE: 
        case MALICIOUS_CAMOUFLAGED: 
        case MALICIOUS_SPY: 
        case MALICIOUS_HIDDEN_A: 
        case MALICIOUS_HIDDEN_B: 
        case MALICIOUS_HIDDEN_SPY: 
          if (collectiveNeighbors_.contains(new Integer(peer.peerID())))
            opinion = 1.0; // judge based on malicious acquaintance
          else if (!authenticFile) 
            opinion = 1.0;
          else
            opinion = -1.0;
          break;
      }
    }
    return opinion;
  }

  /**
   * Initializes uptime ratio and query activity ratio for all specified peers.
   * @param type 0 = always up, 1 = random, 
   *             2 = always up yet no query responses
   */
  public void activity(int activity) 
  {
    switch(activity) 
    {
      case PeerBehavior.ACTIVITY_ALWAYSUP: 
      {
        uptime_        = 1;
        queryActivity_ = 1;
        break;
      }
      case PeerBehavior.ACTIVITY_RANDOM: 
      {
        if (Math.random() < 0.05) {            
          uptime_ = 0.95 + Math.random() * 0.05;
        } else {
          double rand = Math.random();
          if (rand < 0.4)
            uptime_ = Math.random() * 0.075;
          else if (rand < 0.8)
            uptime_ = 0.075 + Math.random() * 0.325;
          else if (rand < 0.95)
            uptime_ = 0.4 + Math.random() * 0.55;
        }    
        // randomly pick query activity
        double fraction = Math.ceil(Math.random() * 8) * 0.1;
        queryActivity_ = fraction > 0 ? fraction : 0.1;
        break;
      }
      case PeerBehavior.ACTIVITY_ALWAYSUPNOQUERIES: 
      {
        // peer does not issue any queries
        uptime_  = 1;
        queryActivity_ = 0;
        break;
      }
    }
  }
}
