/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer.impl;

import java.util.*;
import java.io.*;

import qcsim.peer.*;
import qcsim.peer.content.*;
import qcsim.util.*;
import qcsim.net.util.*;

 /*
  * Title: Query Cycle Simulator
  * Description: P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * Represents a peer's behavior during a simulation.  The peer will
  * use an object of this type to make decisions regarding its behavior.
  * @author unascribed
  * @version 1.0
  */

public class PeerBehaviorRTR extends PeerBehaviorImpl {

    public static final int FILE_PRICE = 1;
    public static final double EXACT_MATCH_SCORE = 1;
    public static final double APPROX_MATCH_SCORE = .01;

    public static final int FILTER_NONE     = 0;
    public static final int FILTER_CAT      = 1;
    public static final int FILTER_FILE     = 2;
    public static final int FILTER_SUBCAT   = 3;

    //------------------------
    // Behavior Parameters!

    // Describes the level of filtering at which a peer advertises content
    protected int filterLevel_;

    // The peer's estimate on what fraction of all possible matches will
    // be returned.  SMALLER is more OPTIMISTIC
    protected double fracReturn_;

    // The peer's estimate on the resell value of an RTR
    // SMALLER is more CONSERVATIVE
    protected double [] resellFudgeFactor_;

    // The peer's estimate on the probability that a neighbor will buy
    // an appropriate priced RTR (e.g., a peer won't buy if he's already
    // seen it before)
    protected double probBuyFudgeFactor_[];

    // The bargain hunter's factor
    protected double bargainFactor_;


    /**
     * Constructor
     * @param manager The manager of the peer corresponding to this behavior
     * @param behavior Initial behavior of the peer.
     */
    public PeerBehaviorRTR(Peer peer, int type) {
	// @@@ TODO: for now, behavior is hard-coded.  Later, we need to
	// feed in behavior vectors, and possibly combine existing ones
	super(peer, type);

	probBuyFudgeFactor_ = new double[PeerMsg.MAX_TTL];

	// @@@ for now
	//type_ = NO_FORWARD_PEER;
	type_ = ALL_FORWARD_PEER;
	//type_ = COMPETITIVE_FORWARD_PEER;

	// @@@ TODO: have a way of setting parameters!
	filterLevel_ = FILTER_FILE;

	bargainFactor_ = 1;

	fracReturn_ = .5;
	for (int i=0; i < PeerMsg.MAX_TTL; i++) 
	    probBuyFudgeFactor_[i] = (double) i / PeerMsg.MAX_TTL;
    }
    
    /**
     * Will determine if peer should respond to a query
     * @param msg The query message
     */
    public boolean respond(PeerMsgQuery msg)
    {
	// check if peer is up to answer queries
	if (Math.random() >= uptime_) {
	    System.err.println("Peer uptime is "+uptime_);
	    return false;
	}
	
	// @@@ we will respond if any file with a matching category exists
	if (peer_.content() == null) {
	    System.err.println("peer content is null?");
	    System.exit(1);
	}
	
	Content content = peer_.content().findCategory(msg.category());
	return (content != null && (content.files() > 0));
    }
    
    /**
     * Will determine the file upload status.
     * For RTRs, this is always good
     */
    public double uploadStatus(PeerMsgRequestDownload msg) {
	return Content.GOOD_FILE;
    }
    
    /**
     * Will determine the value a peer should use to update it's local trust
     * @param peer the peer that was downloaded from.
     * @param authenticFile true if the file download successful.
     */
    public double opinion(Peer peer, boolean authenticFile) {
	// For RTRs, this function is called when one peer downloads from
	// another, by the peer providing the download.  This is the buyer's
	// reputation, so it is always increased by 1
	return 1;
    }
    
    /**
     * Initializes uptime ratio and query activity ratio for all specified peers.
     * @param type 0 = always up, 1 = random, 
     *             2 = always up yet no query responses
     */
    public void activity(int activity) 
    {
	// @@@ possibly make peers up all the time? since it's an economic
	// system, peers have more of an incentive to stay up...
	//super.activity(activity);
	
	uptime_        = 1;
	
	// randomly pick query activity
	double fraction = Math.ceil(Math.random() * 8) * 0.1;
	queryActivity_ = fraction > 0 ? fraction : 0.1;
    }
    
    //=========================================
    // Price Model functions
    //=========================================
    public boolean buyRTR(PeerMsgOffer msg) {
	double value = 0;
	int numNeighbors = peer_.neighbors();
	for (int i=0; i < numNeighbors; i++) {
	    PeerRTR neigh = (PeerRTR) peer_.neighbor(i);
	    if (sellRTR(msg, neigh)) {
		double price = priceRTR(msg, neigh);
		value += price * probBuyRTR(msg, neigh);
	    }
	}

	double myUploadIncome = uploadIncome(msg);
	value += myUploadIncome;
	double theirPrice = msg.price();
	if (value / bargainFactor_ >= theirPrice)
	    return true;
	else {
	    double approx = filterNumApproxMatch(msg);
	    double exact = filterNumExactMatch(msg);
	    ((PeerManagerRTR)peer_.manager()).numPriceRejects++;
	    return false;
	}
    }
    
    public boolean sellRTR(PeerMsgRTR msg, Peer neighbor) {
	if (msg.ttl() == 1) // this message will be dropped
	    return false;

	if (msg.buyerRep() < 0) // no forwarding of bad peer messages?
	    return false;

	if (type_ == NO_FORWARD_PEER)
	    return false;
	if (type_ == COMPETITIVE_FORWARD_PEER)
	    return (peer_.content().findCategory(msg.category()) == null);
	if (type_ == ALL_FORWARD_PEER)
	    return true;

	// sell if matches filters
	PeerBehaviorRTR b = (PeerBehaviorRTR)neighbor.behavior();
	boolean bSell =  (b.filterMatch(msg));
	if (!bSell) {
	    ((PeerManagerRTR)peer_.manager()).numFilterRejects++;
	}
	return bSell;
    }


    public double priceRTR(PeerMsgRTR msg, Peer neighbor) {
      PeerBehaviorRTR b = (PeerBehaviorRTR)neighbor.behavior();
		
      double price = uploadIncome(msg, neighbor);
      // @@@ the below should be 
      price *= resellFudgeFactor_[b.filterLevel()]; 
      return price;
    }

    // This should be a function of age...
    // This is one of the variables
    // Expects the TTL of the offer BEFORE it is sent
    public double probBuyRTR(PeerMsgRTR msg, Peer neighbor) {
      return probBuyFudgeFactor_[msg.ttl()];
    }

    //=========================================
    // Price Model helpers
    //=========================================
    public double uploadIncome(PeerMsgQuery msg) {
	Content content = peer_.content().findCategory(msg.category());
	if (content == null)
	    return 0;
	return uploadIncome(content.files(),
			    content.containsFile(msg.file()) ? 1:0,
			    msg.category(), msg.file());
    }

    public double uploadIncome(PeerMsgQuery msg, Peer neighbor) {
	PeerBehaviorRTR b = (PeerBehaviorRTR)neighbor.behavior();

	return uploadIncome(b.filterNumApproxMatch(msg),
			    b.filterNumExactMatch(msg),
			    msg.category(), msg.file());
    }

    // returns the expected income from uploads
    public double uploadIncome(double numApprox, double numExact,
			       int category, int file) {
	// first, add all my own scores
	double myScore = numApprox * APPROX_MATCH_SCORE;
	myScore += numExact * (EXACT_MATCH_SCORE - APPROX_MATCH_SCORE);

	// now, calculate the total score of all matches out there
	PeerContentManager pcm = peer_.manager().contentManager();
	int numFiles = pcm.getTotalNumFiles();
	double totalScore = pcm.getCategoryPopularity(category) 
	    * fracReturn_ * numFiles * APPROX_MATCH_SCORE;
	totalScore += pcm.getFilePopularity(category, file) 
	    * fracReturn_ * numFiles * 
	    (EXACT_MATCH_SCORE - APPROX_MATCH_SCORE); 
	
	// @@@ TODO: how to translate reputation into score?
	return (myScore / totalScore * FILE_PRICE);
    }
    
  public int filterLevel() {return filterLevel_;}

  public void filterLevel(int level) {filterLevel_ = level;}

  public void resellFudgeFactor(double[] fudge) {resellFudgeFactor_ = fudge;}

  public int resellFudgeFactors()
  {
    return resellFudgeFactor_.length;
  }

  public double resellFudgeFactor(int i)
  {
    return resellFudgeFactor_[i];
  }

  public void fracReturn(double frac) {fracReturn_ = frac;}

  public double fracReturn() { return fracReturn_; }

  public void probBuy(double [] probBuy) 
  {
    for (int i = 0; i < probBuy.length; i++)
     probBuyFudgeFactor_[i] = probBuy[i];
  }

  public int buyFudgeFactors()
  {
    return probBuyFudgeFactor_.length;
  }

  public double buyFudgeFactor(int i)
  {
    return probBuyFudgeFactor_[i];
  }

    // returns true if the message matches our filters
    // recall that we only need to match approximately
    public boolean filterMatch(PeerMsgQuery msg) 
    {
          // do we have anything in our content?
          if ( peer_.content() == null) {
              System.err.println("Not modeling peers with no content");
              System.exit(1);
          }

          switch(filterLevel_) {
	case FILTER_NONE:
	    return true;

	case FILTER_CAT:
	case FILTER_FILE:
	    Content content = peer_.content().findCategory(msg.category());
	    return (content != null && (content.files() > 0));

	case FILTER_SUBCAT:
	    System.err.println("Not yet implemeneted subcat filter");
	    System.exit(1);

	    /*
	case FILTER_FILE:
	    Content cont = peer_.content().category(msg.category());
	    return (cont != null && (cont.containsFile(msg.file())));
	    */
	}
	return false;
    }

    public double precomputeNumFilesInCategory(int category) 
    {
	// @@@ this can be precomputed!
	PeerContentManager pcm = peer_.manager().contentManager();
	double totalPop = 0;
	int numFiles = 0;
	for (int i = 0; i < peer_.content().categories(); i++) {
	    Content content = (Content) peer_.content().category(i);
	    //  @@@ can people lie by falsely reporting categories or
	    // files? i.e. decreasing the number of approx matches?
	    // If content.file == 0, then we shouldn't be reporting 
	    // this category.
	    if (content.files() == 0) 
		continue;
	    
	    totalPop += pcm.getCategoryPopularity(content.category());
	    numFiles += content.files();
	}

	if (numFiles == 0)
	    return 0;

	double myCategoryPop = pcm.getCategoryPopularity(category);
	return (numFiles * myCategoryPop / totalPop);
    }

    // Estimates the number of approximate matches for the query
    // given our current filter level
    // Notice although we can know the exact the number of matches, this 
    // function is actually called by some other peer, who does not know
    // our exact number of matches, but only our filters
    public double filterNumApproxMatch(PeerMsgQuery msg) {
      double guess, guess2;
      switch(filterLevel_) {
        case FILTER_NONE:
        case FILTER_CAT:
          guess = (peer_.content().totalFiles() * 
                   peer_.manager().contentManager().getCategoryPopularity(msg.category())); 

          guess2 = (precomputeNumFilesInCategory(msg.category()));

          // @@@ debug
          if (guess >= guess2 && guess > 0) {
            System.out.println("huh?");
          }
          ((PeerManagerRTR)peer_.manager()).totalApproxDiff += (guess2 - guess);
          ((PeerManagerRTR)peer_.manager()).numApproxGuesses++;

          if (filterLevel_ == FILTER_NONE)
            return guess;
          return guess2;

          case FILTER_SUBCAT:
            System.err.println("Not yet implemeneted subcat filter");
            System.exit(1);
          case FILTER_FILE:
            Content content = peer_.content().findCategory(msg.category());
            return (content == null) ? 0 : content.files();
      }
      return 0;
    }

    // Estimates the number of exact matches for the query
    // given our current filter level
    // Notice although we can know the exact the number of matches, this 
    // function is actually called by some other peer, who does not know
    // our exact number of matches, but only our filters
    public double filterNumExactMatch(PeerMsgQuery msg) {
	double guess = 0, guess2 = 0;
	switch(filterLevel_) {
	case FILTER_NONE:
	case FILTER_CAT:
	    guess = peer_.content().totalFiles() * 
		peer_.manager().contentManager()
		.getFilePopularity(msg.category(), msg.file());
	    if (guess > 1)
		return 1;

	    PeerContentManager pcm = peer_.manager().contentManager();
	    double ratio = 
		pcm.getFilePopularity(msg.category(), msg.file()) / 
		pcm.getCategoryPopularity(msg.category());
	    double nf = precomputeNumFilesInCategory(msg.category());
	    guess2 = ratio * nf;
	    if (guess2 > 1)
		return 1;

	    // @@@ debug
	    if (guess >= guess2 && guess > 0) {
		System.out.println("huh?");
	    }
	    ((PeerManagerRTR)peer_.manager()).totalExactDiff += (guess2 -
								  guess);
	    ((PeerManagerRTR)peer_.manager()).numExactGuesses++;
	    if (filterLevel_ == FILTER_NONE)
		return guess;
	    return guess2;

	case FILTER_SUBCAT:
	    System.err.println("Not yet implemeneted subcat filter");
	    System.exit(1);

	case FILTER_FILE:
	    Content content = peer_.content().findCategory(msg.category());
	    if (content != null && content.containsFile(msg.file()))
		return 1;
	    return 0;
	}
	return 0;
    }
}
