/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer.impl;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;


import qcsim.peer.*;
import qcsim.util.*;

 /*
  * Title: Query Cycle Simulator
  * Description: P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * Provides a GUI description of a peer and its downloads.
  * @author unascribed
  * @version 1.0
  */

public class PeerDescriptionImpl implements PeerDescription, Serializable
{
  private static final int side = 7;

  // This field will not be serialized to disk.
  private PeerImpl peer_;

  /**
   * The x position on the display
   */
  private int xpos_;

  /**
   * The y position on the display
   */
  private int ypos_;

  /** Display Height */
  private int height_;

  /** Display Width */
  private int width_;

  /**
   * Constructor
   */
  public PeerDescriptionImpl() {}

  /**
   * Constructor
   * @param peer the peer to describe.
   */
  public PeerDescriptionImpl(PeerImpl peer)
  {
    peer_          = peer;
  }

  /**
   * String that represents the object this desription describes.
   * @return String containing the peer id.
   */
  public String toString()
  {
    return new String("Peer "+ peer_.peerID());
  }

  /**
   * Set the dimension of the display on which to paint the peer.
   * @param dim The display dimension.
   * @param random A random number used to determine a peer's y corrdinate.
   */
  public void displayDescription(Dimension dim, double rand)
  {
    width_  = (int)(dim.getWidth() * 0.9);
    int offset = (int)(dim.getWidth() * 0.05);

    xpos_  = (int)(width_ * (peer_.peerID() + 1)) / peer_.manager().peers();
    xpos_ += offset;

    height_ = (int)(dim.getHeight() * 0.90);
    ypos_  = (int)(rand * height_);
    ypos_ += 2*side;
  }

  public void move(PeerDescription neighbor, boolean away)
  {
    if (away)
    {
      if (position().getY() < neighbor.position().getY())
        ypos_ -= (int)(ypos_ * 0.4);
      else
        ypos_ += (int)((height_ - ypos_) * 0.4); 
    }
    else
    {
      int nypos = (int) neighbor.position().getY();
      if (position().getY() < neighbor.position().getY())
        ypos_ += (int)((nypos - ypos_) * 0.1);
      else
        ypos_ -= (int)((ypos_ - nypos) * 0.1); 
    }
  } 

  /**
   * Does this peer's description contain the x and y coordinate.
   * @return true if the description contains the x, y
   */
  public boolean contains(int x, int y) 
  { 
    Rectangle2D shape = new Rectangle2D.Double((double)xpos_, (double)ypos_, 
                                               (double)side, (double)side);
    return shape.contains(x, y); 
  }

  /**
   * Get a list of download descriptions.
   * @return vector of download descriptions.
   */
  public Vector downloadDescriptions()
  {
    Vector              downloads = new Vector();
    DownloadDescription download  = null;

    // Add bad downloads first.
    for (int i = 0; i < peer_.badResponses(); i++)
    {
      Peer remote = peer_.badResponse(i);
      download = new DownloadDescription(remote, peer_, Color.red);
      downloads.addElement(download);
    }

    // And finally the good download.
    if (peer_.goodResponse() != null)
    {
      Peer remote = peer_.goodResponse();
      download = new DownloadDescription(remote, peer_, Color.green);
      downloads.addElement(download);
    }

    return downloads;
  }

  /**
   * This will draw the peer with the specified color.
   * @param g The graphics to draw on.
   * @param color The color of the peer.
   */
  public void draw(Graphics g, Color color)
  {
    synchronized( g )
    {
      g.setColor(Color.white);
      g.clearRect((int)(position().getX() - (side/2)), 
                  (int)(position().getY() - (side/2)), side, side);
      g.drawRect((int)(position().getX() - (side/2)), 
                 (int)(position().getY() - (side/2)), side, side);
      g.fillRect((int)(position().getX() - (side/2)), 
                 (int)(position().getY() - (side/2)), side, side);
      g.setColor(color);
      g.drawRect((int)(position().getX() - (side/2)), 
                 (int)(position().getY() - (side/2)), side, side);
      g.fillRect((int)(position().getX() - (side/2)), 
                 (int)(position().getY() - (side/2)), side, side);
    }
  }

  /**
   * Will draw the links connecting this peer to its neighbor.
   * @param color Used to color the link.
   */
  public void drawLinks(Graphics g)
  {
    for (int i = 0; i < peer_.neighbors(); i++)
    {
      Peer    neighbor = peer_.neighbor(i);
      Point2D start    = neighbor.description().position();

      synchronized (g)
      {
        if (peer_.behavior().type() >= PeerBehavior.MALICIOUS_HIDDEN_A ||
            neighbor.behavior().type() >= PeerBehavior.MALICIOUS_HIDDEN_A)
          g.setColor(Color.black);
        else if (peer_.behavior().type() >= PeerBehavior.MALICIOUS_PEER ||
            neighbor.behavior().type() >= PeerBehavior.MALICIOUS_PEER)
          g.setColor(Color.red);
        else g.setColor(Color.green);

        g.drawLine((int)start.getX(), (int)start.getY(), 
                   (int)position().getX(), (int)position().getY());
      }
    }

    if (peer_.behavior().type() != PeerBehavior.MALICIOUS_COLLECTIVE) return;

    float dash[] = {10.0f};
    for (Enumeration enum = peer_.behavior().collectiveNeighbors().elements(); 
         enum.hasMoreElements(); )
    {
      int  peerID   = ((Integer)enum.nextElement()).intValue();
      Peer neighbor = peer_.manager().peer(peerID);

      Point2D start = neighbor.description().position();
      synchronized (g)
      {
        Graphics2D g2    = (Graphics2D) g;
        Stroke oldStroke = g2.getStroke();
        g2.setStroke(new BasicStroke(1.0f, BasicStroke.CAP_BUTT,
                                     BasicStroke.JOIN_MITER,
                                     10.0f, dash, 0.0f));

        g.setColor(Color.red);
        g.drawLine((int)start.getX(), (int)start.getY(), 
                   (int)position().getX(), (int)position().getY());
        g2.setStroke(oldStroke);
      }
    }
  }

  /**
   * Funtion will return the color that the peer should use to paint itself.
   * @param colorTrust If true then base color off of trust value.
   */
  public Color color(boolean colorTrust)
  {
    Color color = Color.black;

    if (!colorTrust)
    {
      if (peer_.behavior().type() < PeerBehavior.FREERIDER_PEER)
        color = Color.black;
      else if (peer_.behavior().type() == PeerBehavior.HIGHLY_TRUSTED_PEER)
        color = Color.green;
      else if (peer_.behavior().type() == PeerBehavior.FREERIDER_PEER)
        color = Color.blue;
      else if (peer_.behavior().type() == PeerBehavior.MALICIOUS_CAMOUFLAGED)
        color = Color.orange;
      else if (peer_.behavior().type() == PeerBehavior.MALICIOUS_SPY)
        color = Color.pink;
      else if (peer_.behavior().type() == PeerBehavior.MALICIOUS_HIDDEN_A)
        color = Color.magenta;
      else if (peer_.behavior().type() == PeerBehavior.MALICIOUS_HIDDEN_B)
        color = Color.magenta;
      else if (peer_.behavior().type() == PeerBehavior.MALICIOUS_HIDDEN_SPY)
        color = Color.magenta;
      else 
        color = Color.red;
    }
    else
    {
      Integer cycle = new Integer(peer_.manager().manager().cycle()-1);
      float trust = 0;
      if (cycle.intValue() > 0)
        trust = (float)peer_.globalTrust(cycle);

      float normalTrust = 0.3f * (1.0f/((float)peer_.manager().peers()));

      if (trust > 1.0)
      {
        System.err.println("ERROR: GLOBAL TRUST VALUE > 1.0: -> "+ trust);
        System.exit(0);
      }

      if (trust <= normalTrust)
      {
        float shade      = 1.0f;
        float totalPeers = (float)peer_.manager().peers();

        for (float i = 1.0f; i < 20 && (i*normalTrust)/20.0f < trust; i++) 
          shade -= 0.05f;
        color = new Color(shade, 0.0f, 0.0f, shade); 
      }
      else if (trust > normalTrust)
      {
        float shade = 0.1f;

        for (float i = 1.0f; i < 19 && (i*normalTrust < trust); i++)
          shade += 0.05f;
        color = new Color(0.0f, shade, 0.0f, shade); 
      }
    }
    return color;
  }

  /**
   * The current position on the display.  This can be set by
   * calling displayDimension(..)
   * @return The centered point of the description's location
   */
  public Point2D position()
  {
    return new Point2D.Double(xpos_, ypos_);
  }

}
