/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer.impl;

import java.util.*;
import java.io.*;

import qcsim.peer.*;
import qcsim.peer.content.*;
import qcsim.net.*;
import qcsim.net.util.*;

 /*
  * Query Cycle Simulator
  * P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * This object contains all information associated with a peer in
  * the network.
  * @author unascribed
  * @version 1.0
  */

public class PeerImpl implements Peer, Serializable
{
  protected static final double BAD_CONNECTION  = -5.0;
  protected static final double GOOD_CONNECTION =  5.0;

  protected static final double BAD_PEER  = -1.0;
  protected static final double GOOD_PEER =  5.0;

  /** Peer debugging level. */
  protected int debugLevel_;

  /**
   * peer identifier
   */
  protected int peerID_;

  /**
   * the manager that created this peer.
   */
  protected PeerManager manager_;

  /**
   * This peer's Behavior
   */
  protected PeerBehavior behavior_;

  /**
   * This peer's Content
   */
  protected PeerContent content_;

  /**
   * A description of this peer.
   */
  protected PeerDescription description_;

  /**
   * List of descriptions describing this peer's neighbors
   */
  protected Vector neighbors_;

  /**
   * list of responses to a query
   */
  protected Vector queryResponses_;

  /**
   * number of good download attempts in different modes
   */
  protected Vector goodDownloads_;

  /**
   * number of inauthentic download attempts in different modes
   */
  protected Vector inauthenticDownloads_;

  /**
   * number of not answered queries
   */
  protected Vector voidDownloads_;

  /**
   * number of entirely failed downloads
   */
  protected Vector failedDownloads_;

  /**
   * number of good uploads
   */
  protected Vector goodUploads_;

  protected Vector authenticResponses_;

  protected Vector totalResponses_;

  /**
   * number of inauthentic uploads
   */
  protected Vector inauthenticUploads_;

  /**
   * Number of query response attempts
   */
  protected Vector queryResponseAttempts_;

  /**
   * History of bad responses to issued queries
   * This is represented, for each cycle, as a
   * Vector of PeerDescriptions
   * of the Peers that gave a bad response to a query.
   */
  protected Vector badResponses_;

  /**
   * History of good responses to issued queries
   * For each cycle we keep the PeerDescription of the
   * Peer that gave the good response to a query.
   */
  protected Vector goodResponse_;

  /**
   * The total number of disconnets this peer has done.
   */
  protected int disconnects_;

  /**
   * Local Trust values
   */
  protected Map localTrust_;

  /**
   * Connection opinion value
   */
  protected Hashtable connectionOpinions_;

  /**
   * Global trust values for each cycle.
   */
  protected Vector globalTrust_;

  protected double oldTrust_;

  /**
   * The messages that this peer has received.
   */
  protected Hashtable messages_;

  /** The shortest path from this peer to all other peers */
  protected ArrayList shortestPath_;

  protected int shortestPathCycle_;

  /** Used to compute shortest path length */
  int distance_;

  protected int ttl_;

  /**
   * Constructor.
   * @param manager The manager that created this peer.
   * @param peerID peer ID
   * @param behavior The peer's behavior.
   */
  public PeerImpl(PeerManager manager, int peerID, int behavior)
  {
    debugLevel_     = 0;
    peerID_         = peerID;
    manager_        = manager;
    behavior_       = new PeerBehaviorImpl(this, behavior);
    content_        = null;
    description_    = new PeerDescriptionImpl(this);
    ttl_            = PeerMsg.MAX_TTL;


    neighbors_             = new Vector();
    queryResponses_        = new Vector();
    goodDownloads_         = new Vector();
    inauthenticDownloads_  = new Vector();
    voidDownloads_         = new Vector();
    failedDownloads_       = new Vector();
    goodUploads_           = new Vector();
    authenticResponses_    = new Vector();
    totalResponses_        = new Vector();
    inauthenticUploads_    = new Vector();
    queryResponseAttempts_ = new Vector();
    goodResponse_          = new Vector();
    badResponses_          = new Vector();
    localTrust_            = new HashMap();
    connectionOpinions_    = new Hashtable();
    globalTrust_           = new Vector();
    messages_              = new Hashtable();
    disconnects_           = 0;
    shortestPath_          = null;
    shortestPathCycle_     = -1;

    neighbors_.addElement( new Vector() );
    goodDownloads_.addElement( new Integer(0) );
    inauthenticDownloads_.addElement( new Integer(0) );
    voidDownloads_.addElement( new Integer(0) );
    failedDownloads_.addElement( new Integer(0) );
    goodUploads_.addElement( new Integer(0) );
    authenticResponses_.addElement( new Integer(0) );
    totalResponses_.addElement( new Integer(0) );
    inauthenticUploads_.addElement( new Integer(0) );
    queryResponseAttempts_.addElement( new Integer(0) );
    goodResponse_.addElement( new Integer(-1) );
    badResponses_.addElement( new Vector() );
  }

  /**
   * Constructor.
   * @param manager The manager that created this peer.
   * @param peerID peer ID
   * @param behavior The peer's behavior.
   */
  public PeerImpl(int debug, PeerManager manager, int peerID, int behavior)
  {
    this(manager, peerID, behavior);
    debugLevel_ = debug;
  }

  public String toString()
  {
    return Integer.toString(peerID());
  }

  /**
   * Peer debug level
   * @return Debug level for this peer.
   */
  protected int debugLevel()
  {
    return debugLevel_;
  }

  /**
   * get peer ID
   * @return peer id.
   */
  public int peerID()
  {
    return peerID_;
  }

  public void ttl(int ttl)
  {
    ttl_ = ttl;
  }

  /**
   * The description of this peer.
   * @return the peer's description object.
   */
  public PeerDescription description()
  {
    return description_;
  }

  /**
   * get peer content category
   * @return the peer's content
   */
  public PeerContent content()
  {
    return content_;
  }

  /**
   * Set peer content category
   * @param content the content
   */
  public void content(PeerContent content)
  {
    content_ = content;
  }

  /**
   * Get peer's behavior for a given scenario
   * @param scenario the scenario
   */
  public PeerBehavior behavior()
  {
    return behavior_;
  }

  public PeerManager manager()
  {
    return manager_;
  }

  protected int cycle()
  {
    return manager().manager().cycle();
  }

  /**
   * The number of neighbors this peer currently has.
   * @return number of neighbors.
   */
  public int neighbors()
  {
    if (neighbors_ == null) return 0;

    Vector neighbors = (Vector) neighbors_.elementAt(cycle());
    return neighbors.size();
  }

  public Peer neighbor(int i)
  {
    Vector neighbors = (Vector) neighbors_.elementAt(cycle());

    if (neighbors != null && i < neighbors.size())
    {
      int id = ((Integer)neighbors.elementAt(i)).intValue();
      return manager().peer( id );
    } else return null;
  }

  public Peer isNeighbor(int peerID)
  {
    for (int i = 0; i < neighbors(); i++)
    {
      if (neighbor(i).peerID() == peerID)
        return neighbor(i);
    }
    return null;
  }

  /**
   * Add neighbor peer.
   * @param peer neighbor peer to be added
   * @return false if peer is not neighbor of this peer
   */
  public void addNeighbor(Peer peer)
  {
    Vector  neighbors = (Vector) neighbors_.elementAt(cycle());
    Integer id        = new Integer(peer.peerID());

    if (!neighbors.contains(id))
    {
      neighbors.addElement(id);
      // description().move(peer.description(), false);
    }

  }

  /**
   * Remove neighbor peer.
   * @param peer neighbor peer to be removed
   * @return false if peer is not neighbor of this peer otherwise true.
   */
  public void removeNeighbor(Peer peer)
  {
    Vector  neighbors = (Vector) neighbors_.elementAt(cycle());
    Integer id        = new Integer(peer.peerID());

    if (neighbors.contains(id))
    {
      neighbors.remove(id);
      // description().move(peer.description(), true);
    }

  }

  public double clusterCoefficient()
  {
    if (neighbors() < 2) return 0;

    double clique = ((neighbors() * (neighbors() - 1)) / 2.0);
    double actual = 0;

    for (int i = 0; i < neighbors(); i++)
    {
      Peer neighbor = neighbor(i);
      for (int j = 0; j < neighbors(); j++)
      {
        if (j == i)
          continue;
        else if (neighbor.isNeighbor(neighbor(j).peerID()) != null)
          actual++;  // neighbor i is also a nieghbor of j.
      }
    }
    actual /= 2.0;

    return actual / clique;
  }

  public double pathLengthToMalicious()
  {
    ArrayList paths = shortestPaths();
    double    sum   = 0.0;
    int       cnt   = 0;

    for (int i = 0; i < manager().peers(); i++)
    {
      if (manager().peer(i).behavior().type() >= PeerBehavior.MALICIOUS_PEER)
      {
        Integer path = (Integer) paths.get(i);
        if (path.intValue() != INF)
        {
          sum += path.doubleValue();
          cnt++;
        }
      }
    }
    return (cnt == 0) ? INF : sum / cnt;
  }

  public ArrayList shortestPaths()
  {
    int cycle = cycle();

    if (shortestPath_ == null)
      shortestPath_ = new ArrayList( manager().peers() );

    if (shortestPathCycle_ != cycle())
    {
      shortestPath_.clear();

      // Initialize
      for (int i = 0; i < manager().peers(); i++)
      {
        if (i == peerID())
          distance_ = 0;
        else
          ((PeerImpl)manager().peer(i)).distance_ = INF;
      }

      Vector peers = new Vector();
      peers.addElement( this );
      for (int i = 0; i < manager().peers() && i < peers.size(); i++)
      {
        PeerImpl peer = (PeerImpl) peers.elementAt(i);

        for (int j = 0; j < peer.neighbors(); j++)
        {
          PeerImpl neighbor = (PeerImpl) peer.neighbor(j);
          if (peer.distance_ + 1 < neighbor.distance_)
          {
            neighbor.distance_ = peer.distance_ + 1;
            peers.addElement( neighbor );
          }
        }
      }

      shortestPath_ = new ArrayList( manager().peers() );

      for (int i = 0; i < manager().peers(); i++)
      {
        int dist = ((PeerImpl)manager().peer(i)).distance_;
        shortestPath_.add(i, new Integer(dist));
      }
      shortestPathCycle_ = cycle();
    }

    return shortestPath_;
  }

  public void localTrust(int peerID, double trust)
  {
    localTrust_.put(new Integer(peerID), new Double(trust));
  }

  public double localTrust(int peerID)
  {
    Double localTrust = (Double)localTrust_.get(new Integer(peerID));
    return (localTrust == null) ? 0.0 : localTrust.doubleValue();
  }

  public void connectionOpinion(int peerID, double opinion)
  {
    connectionOpinions_.put(new Integer(peerID), new Double(opinion));
  }

  public double connectionOpinion(int peerID)
  {
    Double opinion = (Double)connectionOpinions_.get(new Integer(peerID));
    return (opinion == null) ? 0.0 : opinion.doubleValue();
  }

  /**
   * Get the total number good downloads.
   * @return number of good downloads.
   */
  public int goodDownloads()
  {
    Integer value = (Integer)goodDownloads_.elementAt(cycle());
    return value.intValue();
  }

  /**
   * Update the total number of good downloads.
   * @param num number of good downloads.
   */
  public void goodDownloads(int num)
  {
    goodDownloads_.setElementAt(new Integer(num), goodDownloads_.size() - 1);
  }

  /**
   * Get the total number of inauthentic downloads.
   * @return number of inauthentic downloads.
   */
  public int inauthenticDownloads()
  {
    Integer value = (Integer)inauthenticDownloads_.elementAt(cycle());
    return value.intValue();
  }

  /**
   * Get the total number of inauthentic downloads.
   * @return number of inauthentic downloads.
   */
  public int inauthenticDownloads(int cycle)
  {
    Integer value = (Integer)inauthenticDownloads_.elementAt(cycle);
    return value.intValue();
  }

  /**
   * Update the total number of inauthentic downloads.
   * @param num number of inauthentic downloads.
   */
  public void inauthenticDownloads(int cycle, int num)
  {
    inauthenticDownloads_.setElementAt(new Integer(num), cycle);
  }

  /**
   * Get the total number of void downloads.
   * @return number of void downloads.
   */
  public int voidDownloads()
  {
    Integer value = (Integer)voidDownloads_.elementAt(cycle());
    return value.intValue();
  }

  /**
   * Update the total number of void downloads.
   * @param num number of void downloads.
   */
  public void voidDownloads(int num)
  {
    int index = voidDownloads_.size() - 1;
    if (index < 0)
      voidDownloads_.addElement(new Integer(num));
    else
      voidDownloads_.setElementAt(new Integer(num), index);
  }

    public int exactMatches() {return 0;}
    public double totalScore() {return 0;}

  /**
   * Get the total number of failed downloads.
   * @return number of failed downloads.
   */
  public int failedDownloads()
  {
    Integer value = (Integer)failedDownloads_.elementAt(cycle());
    return value.intValue();
  }

  /**
   * Update the total number of failed downloads.
   * @param num number of failed downloads.
   */
  public void failedDownloads(int num)
  {
    int index = failedDownloads_.size() - 1;
    if (index < 0)
      failedDownloads_.addElement(new Integer(num));
    else
      failedDownloads_.setElementAt(new Integer(num), index);
  }

  /**
   * Get the total number of good uploads.
   * @return number of good uploads.
   */
  public int goodUploads()
  {
    Integer value = (Integer)goodUploads_.elementAt(cycle());
    return value.intValue();
  }

  public int totalGoodUploads()
  {
    int total = 0;
    for (int i = 0; i < cycle(); i++)
    {
      Integer value = (Integer)goodUploads_.elementAt(i);
      total += value.intValue();
    }
    return total;
  }

  /**
   * Update the total number of good uploads.
   * @param num number of good uploads.
   */
  public void goodUploads(int num)
  {
    int index = goodUploads_.size() - 1;
    if (index < 0)
      goodUploads_.addElement(new Integer(num));
    else
      goodUploads_.setElementAt(new Integer(num), index);
  }

  /**
   * Get the number of authentic responses.
   * @return number of authentic responses.
   */
  public int authenticResponses()
  {
    Integer value = (Integer)authenticResponses_.elementAt(cycle());
    return value.intValue();
  }

  /**
   * Update the total number of authentic responses.
   * @param num number of authentic responses.
   */
  public void authenticResponses(int num)
  {
    int index = authenticResponses_.size() - 1;
    if (index < 0)
      authenticResponses_.addElement(new Integer(num));
    else
      authenticResponses_.setElementAt(new Integer(num), index);
  }

  /**
   * Get the number of total responses.
   * @return number of total responses.
   */
  public int totalResponses()
  {
    Integer value = (Integer)totalResponses_.elementAt(cycle());
    return value.intValue();
  }

  /**
   * Update the total number of total responses.
   * @param num number of total responses.
   */
  public void totalResponses(int num)
  {
    int index = totalResponses_.size() - 1;
    if (index < 0)
      totalResponses_.addElement(new Integer(num));
    else
      totalResponses_.setElementAt(new Integer(num), index);
  }

  /**
   * Get the total number of inauthentic uploads.
   * @return number of inauthentic uploads.
   */
  public int inauthenticUploads()
  {
    Integer value = (Integer)inauthenticUploads_.elementAt(cycle());
    return value.intValue();
  }

  /**
   * Update the total number of inauthentic uploads.
   * @param num number of inauthentic uploads.
   */
  public void inauthenticUploads(int num)
  {
    int index = inauthenticUploads_.size() - 1;
    if (index < 0)
      inauthenticUploads_.addElement(new Integer(num));
    else
      inauthenticUploads_.setElementAt(new Integer(num), index);
  }

  /**
   * Get the total number of query response attempts.
   * @return number query response attempts.
   */
  public int queryResponseAttempts()
  {
    Integer value = (Integer)queryResponseAttempts_.elementAt(cycle());
    return value.intValue();
  }

  /**
   * Update the total number of query response attempts.
   * @param num The number of query response attempts.
   */
  public void queryResponseAttempts(int num)
  {
    int index = queryResponseAttempts_.size() - 1;
    if (index < 0)
      queryResponseAttempts_.addElement(new Integer(num));
    else
      queryResponseAttempts_.setElementAt(new Integer(num), index);
  }

  /**
   * The total cound of bad responses for the current cycle
   * @return The number of bad responses.
   */
  public int badResponses()
  {
    if (badResponses_.size() == 0) badResponses_.addElement(new Vector());

    Vector responses = (Vector)badResponses_.elementAt(cycle());
    return responses.size();
  }

  /**
   * Add the peer description to the list of bad responses.
   * @param desc The description of the peer.
   */
  public void badResponse(Peer peer)
  {
    if (badResponses_.size() == 0) badResponses_.addElement(new Vector());

    Vector responses = (Vector)badResponses_.elementAt(cycle());
    responses.addElement(new Integer(peer.peerID()));
  }

  /**
   * Get the ith bad response during the current cycle.
   * @param i The index (staring from 0) of the desired description.
   * @return A description of the ith peer that sent a bad response.
   */
  public Peer badResponse(int i)
  {
    if (badResponses_.size() == 0) badResponses_.addElement(new Vector());

    Vector responses = (Vector)badResponses_.elementAt(cycle());
    if (i < 0 || i >= responses.size())
      return null;

    int peerid = ((Integer) responses.elementAt(i)).intValue();
    return manager().peer(peerid);
  }

  /**
   * Add the description of the peer that sent a good response.
   * @param desc The description of the peer.
   */
  public void goodResponse(Peer peer)
  {
    int index = goodResponse_.size() - 1;
    if (index < 0)
      goodResponse_.addElement(new Integer(peer.peerID()));
    else
      goodResponse_.setElementAt(new Integer(peer.peerID()), index);
  }

  /**
   * Get the good response cooresponding to the current cycle.
   * @return The description of the peer to sent a good response.
   */
  public Peer goodResponse()
  {
    Integer id = (Integer) goodResponse_.elementAt(cycle());
    return (id.intValue() < 0) ? null : manager().peer(id.intValue());
  }

  /**
   * Yields trust value computed by this peer.
   * @return trust trust value at peer
   */
  public double globalTrust()
  {
    if (globalTrust_.size() <= cycle()) return 0.0;

    Double value = (Double)globalTrust_.elementAt(cycle());
    return value.doubleValue();
  }

  public double globalTrust(Integer cycle)
  {
    if (cycle.intValue() < 0 ||
        globalTrust_.size() <= cycle.intValue())
      return 0.0;

    Double value = (Double)globalTrust_.elementAt(cycle.intValue());
    return value.doubleValue();
  }

  /**
   * Set global trust for a given cycle
   * @param trust the global trust value
   */
  public void globalTrust(double trust)
  {
    globalTrust_.setElementAt(new Double(trust), cycle());
  }

  public int disconnects()
  {
    return disconnects_;
  }

//=============================================================

  /**
   * Sort a vector of peer ids based on local trust from lowest to highest.
   * @return local trust sorted, vector of peer ids.
   */
  protected Vector sortLocalTrust()
  {
    Vector  peers  = new Vector();
    HashMap local  = (HashMap) ((HashMap)localTrust_).clone();
    Integer peerID = null;
    Integer min    = null;
    double  trust  = 0;

    while (local.size() > 0)
    {
      for (Iterator iter = local.keySet().iterator(); iter.hasNext(); )
      {
        peerID = (Integer) iter.next();
        trust  = localTrust(peerID.intValue());

        if (min == null || trust < localTrust(min.intValue()))
          min = peerID;
      }
      peers.addElement( min );
      local.remove( min );
      min = null;
    }

    return peers;
  }

  protected Peer pickNewNeighbor()
  {
    Vector  sort     = sortLocalTrust();
    Peer    neighbor = null;

    // Choose known neighbor with good local trust.
    if (behavior().type() < PeerBehavior.MALICIOUS_PEER)
    {
      for (int i = sort.size() - 1; i >= 0; i--)
      {
        int id = ((Integer)sort.elementAt(i)).intValue();

        if (localTrust(id) < 1.0)
          break;
        else if (connectionOpinion(id) < 0.0)
          continue;

        neighbor = isNeighbor(id);
        if (neighbor == null)
        {
          PeerMsgRequestConnect msg = new PeerMsgRequestConnect(peerID(), id);
          if (((Boolean)network().rpc(msg)).booleanValue())
            return manager_.peer( id );
        }
      }
    }
    else
    {
      for (int i = 0; i < sort.size(); i++)
      {
        int id = ((Integer)sort.elementAt(i)).intValue();


        neighbor = isNeighbor(id);
        if (neighbor == null)
        {
          PeerMsgRequestConnect msg = new PeerMsgRequestConnect(peerID(), id);
          if (((Boolean)network().rpc(msg)).booleanValue())
            return manager_.peer( id );
        }
      }
    }

    // Choose random peer to connect.
    Random rand = new Random();
    for (int i = 0; i < 5; i++)
    {
      neighbor = manager().randomPeer(rand);

      if (neighbor.peerID() == peerID() ||
          isNeighbor(neighbor.peerID()) != null)
        continue;
      else if (connectionOpinion(neighbor.peerID()) <= BAD_CONNECTION)
        continue;
      else if (localTrust(neighbor.peerID()) <= BAD_PEER)
        continue;

      PeerMsgRequestConnect pmrc = new PeerMsgRequestConnect(peerID(), neighbor.peerID());
      Boolean connect = (Boolean) network().rpc(pmrc);
      if (!connect.booleanValue())
        continue;  // Request to connect denied.

      return neighbor;
    }
    return null;
  }

  protected Peer pickNeighborToRemove()
  {
    if (neighbors() == 0) return null;

    if (behavior().type() < PeerBehavior.MALICIOUS_PEER)
    {
      Peer lt = null;
      Peer co = null;

      double sumNeighborLocalTrust = 0.0;
      ArrayList sortLT = new ArrayList();

      for (int i = 0; i < neighbors(); i++)
      {
        int id = neighbor(i).peerID();

        sumNeighborLocalTrust += localTrust(id);
        int pos = 0;
        for ( ; pos < sortLT.size(); pos++)
        {
          int pid = ((Integer)sortLT.get(pos)).intValue();
          if (localTrust(pid) > localTrust(id))
            break;
        }
        sortLT.add(pos, new Integer(id));

        if (lt == null || localTrust(id) < localTrust(lt.peerID()))
          lt = neighbor(i);
        if (co == null || connectionOpinion(id)<connectionOpinion(co.peerID()))
          co = neighbor(i);
      }

      if (localTrust(lt.peerID()) <= 0.0)
        return lt;
      if (connectionOpinion(co.peerID()) <= 0.0)
        return co;

      double rand = Math.random();
      double[] dist = new double[sortLT.size()];
      for (int i = 0; i < sortLT.size(); i++)
        dist[i] = localTrust(((Integer)sortLT.get(i)).intValue()) /
                  sumNeighborLocalTrust;

      int pid = 0;
      int i   = dist.length - 1;
      for (double sum = dist[i]; i > 0 && sum < rand; i--, sum += dist[i])
        pid++;
      pid = ((Integer) sortLT.get(pid)).intValue();
      return manager().peer(pid);
    }
    else if (behavior().type() != PeerBehavior.MALICIOUS_HIDDEN_SPY)
    {
      Peer neighbor = null;
      for (int i = 0; i < neighbors(); i++)
      {
        if (neighbor == null ||
            localTrust(neighbor(i).peerID()) > localTrust(neighbor.peerID()))
        {
          if (neighbor(i).behavior().type()!=PeerBehavior.MALICIOUS_HIDDEN_SPY)
            neighbor = neighbor(i);
        }
      }
      return neighbor;
    }

    return null;
  }

  protected void changeNeighbors()
  {
    Peer rn = pickNeighborToRemove();
    if (rn != null)
    {
      network().rpc( new PeerMsgDisconnect(peerID(), rn.peerID()) );
      disconnects_++;

      Peer nn = pickNewNeighbor();
      if (nn != null)
      {
        nn.addNeighbor( this );
        addNeighbor( nn );
      }
    }
  }


// ==================== MESSAGE INTERFACE =============================


  /**
   * Dispatches message to associated message handler.
   * @param msg message received
   */
  public void message(PeerMsg msg)
  {
    if (msg.source() == peerID())
      return;   // this is my message.

    // determine type of message and call corresponding message handler
    if (msg instanceof PeerMsgQuery)
      handlePeerMsgQuery((PeerMsgQuery)msg);
    else if (msg instanceof PeerMsgQueryResponse)
      handlePeerMsgQueryResponse((PeerMsgQueryResponse)msg);
    else if (msg instanceof PeerMsgTrust)
      handlePeerMsgTrust((PeerMsgTrust)msg);
  }

  /**
   * Handles query messages.
   * @param msg        query message
   */
  protected void handlePeerMsgQuery(PeerMsgQuery msg)
  {
    if (messages_.get(msg) != null)
      return;   // have already seen this message.
    messages_.put(msg, msg);

    // forward message to all neighbor peers
    for (int i = 0; i < neighbors(); i++)
      network().send(neighbor(i).peerID(), msg);

    queryResponseAttempts(queryResponseAttempts()+1);

    if (debugLevel() > 1)
      System.out.println("Peer "+peerID()+" processes query.");

    if (behavior().respond(msg))
    {
      PeerMsg msgr = new PeerMsgQueryResponse(peerID(), msg.source(),
                                              msg.category(), msg.file(),
                                              msg.connection(), msg.ttl());
      network().send(msg.source(), msgr);
    }
  }

  /**
   * Handles query response messages.
   * @param msg query message response
   */
  protected void handlePeerMsgQueryResponse(PeerMsgQueryResponse msg)
  {
      /*
    if (messages_.get(msg) != null)
      return;   // have already seen this message.
    messages_.put(msg, msg);
      */

    if (msg.destination() == peerID())
      queryResponse(msg); // push onto list of responses
  }

  /**
   * Handles opinion messages.
   * @param msg opinion message
   */
  protected void handlePeerMsgTrust(PeerMsgTrust msg)
  {
    double current = globalTrust();
    globalTrust(current+msg.weightedTrustValue());
  }

  public Object rpc(PeerMsg msg)
  {
    if (msg instanceof PeerMsgRequestConnect)
      return handleRequestConnect((PeerMsgRequestConnect) msg);
    if (msg instanceof PeerMsgDisconnect)
      return handleRequestDisconnect((PeerMsgDisconnect) msg);
    else if (msg instanceof PeerMsgRequestDownload)
      return handleRequestDownload((PeerMsgRequestDownload) msg);
    else if (msg instanceof PeerMsgRequestTrust)
      return handleRequestTrust((PeerMsgRequestTrust) msg);
    return null;
  }

  protected Boolean handleRequestConnect(PeerMsgRequestConnect msg)
  {
    Double copn = (Double)connectionOpinions_.get(new Integer(msg.source()));

    if (localTrust(msg.source()) < 0)
      return new Boolean(false);
    else if (copn != null && copn.intValue() <= BAD_CONNECTION)
      return new Boolean(false);
    else if (neighbors() > NetworkManager.MAX_CONNECTIONS)
      return new Boolean(false);

    if (neighbors() < NetworkManager.MAX_CONNECTIONS)
      return new Boolean(true);
    else if (neighbors() > 0)
    {
       /* If i'm going to connect to a new peer I must remove a
        * current neighbor.  Make sure that it is worth it.
        */
       Peer neighborToRemove = pickNeighborToRemove();

       if (localTrust(neighborToRemove.peerID()) < localTrust(msg.source()))
         return new Boolean(true);
    }
    return new Boolean(false);
  }

  protected Boolean handleRequestDisconnect(PeerMsgDisconnect msg)
  {
    manager().peer(msg.source()).removeNeighbor( this );
    removeNeighbor( manager().peer( msg.source() ) );
    return new Boolean(true);
  }

  protected Double handleRequestDownload(PeerMsgRequestDownload msg)
  {
    double status = behavior().uploadStatus(msg);
    Peer   src    = manager().peer(msg.source());

    if (behavior().type() < PeerBehavior.MALICIOUS_PEER ||
        src.behavior().type() < PeerBehavior.MALICIOUS_PEER)
    {
      if (status == (double)Content.GOOD_FILE)
        goodUploads(goodUploads()+1);
      else
        inauthenticUploads(inauthenticUploads()+1);
    }

    return new Double( status );
  }

  protected Double handleRequestTrust(PeerMsgRequestTrust msg)
  {
    return new Double(globalTrust());
  }

  protected NetworkManager network()
  {
    return manager().network();
  }

//================================================================
// Peer Manager interface.

  /**
   * Initialize a new cycle.
   */
  public void newCycle()
  {
    if (neighbors_.size() == cycle())
    {
      int prev;

      neighbors_.addElement(((Vector)neighbors_.lastElement()).clone());

      prev = ((Integer) goodDownloads_.lastElement()).intValue();
      goodDownloads_.addElement(new Integer(0));

      // prev = ((Integer) inauthenticDownloads_.lastElement()).intValue();
      inauthenticDownloads_.addElement(new Integer(0));

      prev = ((Integer) voidDownloads_.lastElement()).intValue();
      voidDownloads_.addElement(new Integer(0));

      prev = ((Integer) failedDownloads_.lastElement()).intValue();
      failedDownloads_.addElement(new Integer(prev));

      // prev = ((Integer) goodUploads_.lastElement()).intValue();
      goodUploads_.addElement(new Integer(0));

      // prev = ((Integer) inauthenticUploads_.lastElement()).intValue();
      inauthenticUploads_.addElement(new Integer(0));

      prev = ((Integer) queryResponseAttempts_.lastElement()).intValue();
      queryResponseAttempts_.addElement(new Integer(prev));

      goodResponse_.addElement(new Integer(-1));
      badResponses_.addElement(new Vector());

      authenticResponses_.addElement(new Integer(0));
      totalResponses_.addElement(new Integer(0));

      queryResponses_.clear();
      messages_.clear();

    }

    Double trust;
    if (cycle() == 0)
    {
      if (behavior().type() == PeerBehavior.HIGHLY_TRUSTED_PEER)
        trust = new Double(1.0/(double)manager().highlyTrustedPeers());
      else
        trust = new Double(0);
      globalTrust_.addElement(trust);
    }
    else
    {
      trust = (Double) globalTrust_.lastElement();
      globalTrust_.addElement(trust);
    }
  }

  public void preSubmitTrust(int iter)
  {
    if (iter == 0)
    {
      if (manager().highlyTrustedPeers() > 0)
      {
        if (behavior().type() == PeerBehavior.HIGHLY_TRUSTED_PEER)
          oldTrust_ = 1.0/(double)manager().highlyTrustedPeers();
        else oldTrust_ = 0.0;
      }
      else oldTrust_ = 1.0/(double)manager().peers();
    }
    else oldTrust_ = globalTrust();

    globalTrust(0);
  }

  /**
   * Tells peer to send it's local trust value to others.
   */
  public void submitTrust()
  {
    PeerMsgTrust msg     = null;

    Set    peers         = localTrust_.keySet();
    int    peerID        = 0;
    double highTrust     = 0.0;
    double normTrust     = 0.0;
    double sumLocalTrust = 0.0;

    if (oldTrust_ == 0) return;

    for (Iterator iter=peers.iterator(); iter.hasNext(); )
    {
      peerID = ((Integer)iter.next()).intValue();
      sumLocalTrust += localTrust(peerID) < 0 ? 0 : localTrust(peerID);
    }

    if (sumLocalTrust > 0)
    {
      for (Iterator iter=peers.iterator(); iter.hasNext(); )
      {
        peerID = ((Integer)iter.next()).intValue();
        if (peerID == peerID()) continue;

        normTrust  = localTrust(peerID) < 0 ? 0 : localTrust(peerID);
        normTrust /= sumLocalTrust;
        normTrust *= oldTrust_;

        if (normTrust > 0)
        {
          // System.err.println("PEER: "+peerID()+" SUBMIT TRUST TO PEER "+
          //                  peerID+" trust "+normTrust);
          msg = new PeerMsgTrust(peerID(), peerID, normTrust);
          network().send(peerID, msg);
        }
      }
    }
    else
    {
       /* if peer does not have any opinion it will default trusts.  */
      if (manager().highlyTrustedPeers() > 0)
      {
        double trust = oldTrust_/(double)manager().highlyTrustedPeers();
        // assign trust to highly trusted nodes
        for (int j = 0; j < manager().peers(); j++)
        {
          peerID = manager().peer(j).peerID();

          if (manager().peer(peerID).behavior().type() ==
              PeerBehavior.HIGHLY_TRUSTED_PEER)
          {
            if (peerID == peerID())
              globalTrust(globalTrust() + trust);
            else
            {
              msg = new PeerMsgTrust(peerID(), peerID, trust);
              network().send(peerID, msg);
            }
          }
        }
      }
      else
      {
        double trust = oldTrust_/(double)manager().peers();
        for (int j = 0; j < manager().peers(); j++)
        {
          peerID  = manager().peer(j).peerID();
          if (peerID == peerID())
          {
            globalTrust(globalTrust() + trust);
          }
          else
          {
            msg   = new PeerMsgTrust(peerID(), peerID, trust);
            network().send(peerID, msg);
          }
        }
      }
    }
  }

  public void postSubmitTrust()
  {
    double trust = globalTrust();

    if (manager().highlyTrustedPeers() != 0)
      trust *= 0.85;
    if (behavior().type() == PeerBehavior.HIGHLY_TRUSTED_PEER)
      trust += 0.15*(1.0/(double)manager().highlyTrustedPeers());
    globalTrust(trust);
  }

// ===================================================================

  /**
   * Randomly issues query based on peer activity.
   * @return true if query issued other false.
   */
  public void query()
  {
    if (behavior().type() == PeerBehavior.MALICIOUS_HIDDEN_SPY) return;

    // Try and maintain at least two connections.
    if (manager().adaptiveTopology() != PeerManager.AT_NONE)
    {
      if (neighbors() < 2)
      {
        Peer neighbor = pickNewNeighbor();
        if (neighbor != null)
        {
          neighbor.addNeighbor( this );
          addNeighbor( neighbor );
        }
      }
    }

    // determine if query will be issued
    if (Math.random() <= behavior().queryActivity())
    {
      issueQuery(ttl_);
      processQueryResponses();
    }
  }

  /**
   * Randomly issues query based on peer interest.
   * Sends out query message and puts peer into inactive state.
   * Peer will query for a file rank in some content category.
   * @param TTL broadcast search time-to-live
   */
  protected void issueQuery(int TTL)
  {
    PeerContentManager pcm = manager_.contentManager();
    int category = (int) (Math.random() * pcm.contentCategories()) + 1;
    int file     = (int) (Math.random() * pcm.uniqueFiles(category));

    if (behavior().type() < PeerBehavior.MALICIOUS_PEER)
    {
      category = content().randomCategory();
      file = content().randomFile(category);
    }

    PeerMsgQuery msg = null;
    // forward message to all neighbor peers
    for (int i = 0; i < neighbors(); i++)
    {
      msg = new PeerMsgQuery(peerID(), PeerMsg.ANY, TTL,
                             category, file, neighbor(i).peerID());
      network().send(neighbor(i).peerID(), msg);
    }

    if (debugLevel() >= 1)
      System.out.println("Peer "+peerID_+" issues query in category "+category
                         +" and file rank "+file);
  }


//========================================================================
// Query responses.

  /**
   * Number of query responses.
   * @return number of query responses.
   */
  protected int queryResponses()
  {
    return queryResponses_.size();
  }

  /**
   * Get message cooresponding to the selector
   * @param msgi message index
   */
  protected PeerMsgQueryResponse queryResponse(int msgi)
  {
    return (PeerMsgQueryResponse) queryResponses_.elementAt( msgi );
  }

  /**
   * Add response message
   * @param msg response message to add.
   */
  protected void queryResponse(PeerMsgQueryResponse msg)
  {
    if (!queryResponses_.contains(msg))
      queryResponses_.addElement( msg );
  }


  /**
   * Process responses to query.
   * @param cycle the cycle
   */
  private void processQueryResponses() 
  {
    if (queryResponses() == 0) 
    {
      // no responses, then we are done.
      voidDownloads(voidDownloads()+1);

      // Change neighbors on void downloads but don't do this too often.
      if (manager().adaptiveTopology() != PeerManager.AT_NONE &&
          voidDownloads() % 3 == 0                            &&
          behavior().type() < PeerBehavior.MALICIOUS_PEER)
        changeNeighbors();

      return;
    }

    boolean oneGood = false;
    for (int i = 0; !oneGood && i < queryResponses(); i++)
    {
      if (manager().peer(queryResponse(i).source()).behavior().type() < 
          PeerBehavior.MALICIOUS_PEER) 
      oneGood = true;
    }
    if (!oneGood) return;

    // trust values of responding peers
    double[] responseTrust         = new double[queryResponses()];
    double   sumResponseLocalTrust = 0.0;
    
    totalResponses(queryResponses());
    for (int i = 0; i < queryResponses(); i++) 
    {
      // get trust value for responding peers
      int peerId   = queryResponse(i).source();

      if (manager().peer(peerId).behavior().type()<PeerBehavior.MALICIOUS_PEER)
        authenticResponses(authenticResponses() + 1);

      if (behavior().responseSelector()==PeerBehavior.RPB_SIMPLEPROBABILISTIC) 
      {
        responseTrust[i]       = localTrust(peerId);
        sumResponseLocalTrust += responseTrust[i];
      }
      else 
      {
        PeerMsgRequestTrust msg = new PeerMsgRequestTrust(peerID(), peerId);
        Double trust = (Double) network().rpc(msg);
        responseTrust[i] = trust.doubleValue();
      }
    }

    if (sumResponseLocalTrust > 0.0)
    {
      for (int i = 0; i < responseTrust.length; i++)
        responseTrust[i] = responseTrust[i] / sumResponseLocalTrust;
    }

    
    // order list based on response trust values
    boolean changed = true;
    while (changed) 
    {
      changed = false;
      for (int i = 0; i < queryResponses() - 1; i++) 
      {
        if (responseTrust[i] < responseTrust[i+1]) 
        {
          double tmpTrust    = responseTrust[i+1];
          responseTrust[i+1] = responseTrust[i];
          responseTrust[i]   = tmpTrust;
          Object tmpResponse = queryResponses_.elementAt(i+1);

          queryResponses_.removeElementAt(i+1);
          queryResponses_.insertElementAt(tmpResponse, i);
          changed = true;
        }
      }
    } // END WHILE changed

    if (debugLevel() > 1) 
    {
      System.err.println("Ordered list of responses at peer "+peerID_+":");
      for (int i = 0; i < queryResponses(); i++) 
      {
        System.err.print(queryResponse(i).source() + " ");
      }
      System.err.println("\n");
    }
                            
    Vector  queryResponses          = new Vector();
    double  normDownloadProbability = 0;
    double  normZeroTrustValuePeers = 0;
    int     numZeroTrustValuePeers  = 0;

    for (int i = 0; i < queryResponses(); i++)
    {
      queryResponses.addElement(queryResponse(i));

      normDownloadProbability += responseTrust[i];
      // count number of peers with trust value 0
      if (responseTrust[i] == 0)
        numZeroTrustValuePeers++;
    }

    // assign certain selection probability to peers with trust value 0
    if (numZeroTrustValuePeers > 0)
    {
      if (normDownloadProbability > 0)
      {
        normZeroTrustValuePeers = (normDownloadProbability * 0.05) /
                                  numZeroTrustValuePeers;
        normDownloadProbability += (normDownloadProbability * 0.05);
      }
      else
      {
        normZeroTrustValuePeers = 1.0 / numZeroTrustValuePeers;
        normDownloadProbability = 1.0;
      }
    }
                    
    double trust = 1.0;
    double  download = Content.BAD_FILE;
    int current=0, category=0, file=0;
    while (download != Content.GOOD_FILE && queryResponses.size() > 0) 
    {
      switch (behavior().responseSelector()) 
      {
        case PeerBehavior.RPB_RANDOM: 
          current = (int)(Math.random() * queryResponses.size());
          break;
        case PeerBehavior.RPB_TRUSTDETERMINISTIC: 
          current = 0;  // This first element.
          break;
        case PeerBehavior.RPB_TRUSTPROBABILISTIC: 
        case PeerBehavior.RPB_SIMPLEPROBABILISTIC: 
          double rand = Math.random() * normDownloadProbability;
          double sum  = 0.0;

          for (current = 0; current < queryResponses.size()-1; current++) 
          {
            if (responseTrust[current] > 0)
              sum += responseTrust[current];
            else
              sum += normZeroTrustValuePeers;
            if (sum >= rand) break;
          }
          break;
        case PeerBehavior.RPB_RESPONSETIME:
          int max = 
            ((PeerMsgQueryResponse)queryResponses.elementAt(0)).responseTime();
          int rt  = -1;

          current = 0;
          for (int i = 1; i < queryResponses.size(); i++) 
          {
            rt=
             ((PeerMsgQueryResponse)queryResponses.elementAt(i)).responseTime();
            if (max < rt) 
              current = i;
          }
          break;
        default:
          System.err.println("Unknown response selector");
          System.exit(1);
          break;
      }

      int remotePeerId = ((PeerMsgQueryResponse)queryResponses.elementAt(current)).source();
      int connection = ((PeerMsgQueryResponse)queryResponses.elementAt(current)).connection();
      category = ((PeerMsgQueryResponse)queryResponses.elementAt(current)).category();
      file = ((PeerMsgQueryResponse)queryResponses.elementAt(current)).file();

      if (debugLevel() >= 1)
        System.out.println("Peer " + peerID_ + " requests non-trusted download from peer " +
                           remotePeerId);

      PeerMsg msg = new PeerMsgRequestDownload(peerID(), remotePeerId, category, file); 

      download = ((Double)network().rpc(msg)).doubleValue();

      // update opinion matrix
      recordDownload(remotePeerId, connection, download);

      // delete responding peer from list
      queryResponses.removeElementAt(current);
    }


    // count number of entirely failed downloads
    if (download != Content.GOOD_FILE) 
    {
      failedDownloads(failedDownloads()+1);
    }
    else if (behavior().type() < PeerBehavior.MALICIOUS_PEER)
    {
      Content  content = content().findCategory(category);
      content.addFile(file);
    }
  }
  

  /** Updates opinion on peer.
   * @param peerID ID of peer on which to update opinion
   * @param connection The connection used to send initial query.
   * @param download success of download
   */
  protected void recordDownload(int peerID, int connection, double download)
  {
    Peer    peer    = manager().peer(peerID);
    Integer peerKey = new Integer(peerID);
    Double  copn    = (Double)connectionOpinions_.get(new Integer(connection));
    Double  popn    = (Double)localTrust_.get(peerKey);

    if (copn == null)
      copn = new Double(0);

    if (popn == null)
      popn = new Double(0);

    // update statistics
    if (download == Content.GOOD_FILE)
    {
      goodDownloads(goodDownloads()+1);
      goodResponse( peer );
    }
    else
    {
      inauthenticDownloads(cycle(), inauthenticDownloads(cycle())+1);
      badResponse( peer );
    }

    double updateAmount = behavior().opinion(peer, download==Content.GOOD_FILE);

    Double peerTrust = new Double(popn.doubleValue() + updateAmount);
    if (peerTrust.intValue() != 0)
      localTrust_.put(peerKey, peerTrust);
    else localTrust_.remove(peerKey);

    copn = new Double(copn.doubleValue() + updateAmount);
    connectionOpinions_.put(new Integer(connection), copn);

    if (manager().adaptiveTopology() != PeerManager.AT_NONE)
    {
      Peer neighbor = null;
      boolean dlt   = false;
      boolean dco   = false;
      boolean clt   = false;
      boolean cco   = false;

      if (manager().adaptiveTopology() == PeerManager.AT_LOCALTRUST ||
          manager().adaptiveTopology() == PeerManager.AT_ALL)
      {
        if (peerTrust.intValue() < 0.0)
          dlt = true;
        if (peerTrust.intValue() > GOOD_PEER)
          clt = true;
      }

      if (manager().adaptiveTopology() == PeerManager.AT_CONNECTIONTRUST ||
          manager().adaptiveTopology() == PeerManager.AT_ALL)
      {
        if (copn.doubleValue() < 0.0)
          dco = true;
        if (copn.intValue() > GOOD_CONNECTION)
          cco = true;
      }

      if (dlt && ((neighbor=isNeighbor(peerID)) != null))
      {
        network().rpc( new PeerMsgDisconnect(peerID(), neighbor.peerID()) );
        disconnects_++;

        Peer newNeighbor = pickNewNeighbor();
        if (newNeighbor != null)
        {
          newNeighbor.addNeighbor( this );
          addNeighbor( newNeighbor );
        }
      }
      else if (dco && ((neighbor=isNeighbor(connection)) != null))
      {
        network().rpc( new PeerMsgDisconnect(peerID(), neighbor.peerID()) );
        disconnects_++;

        Peer newNeighbor = pickNewNeighbor();
        if (newNeighbor != null)
        {
          newNeighbor.addNeighbor( this );
          addNeighbor( newNeighbor );
        }
      }
      else if (((neighbor=isNeighbor(peerID)) == null) && (clt || cco))
      {
        Peer ntr = pickNeighborToRemove();
        if (ntr != null && localTrust(ntr.peerID()) < localTrust(peerID))
        {
          PeerMsg pmrc = new PeerMsgRequestConnect(peerID(),peerID);
          if (((Boolean)network().rpc(pmrc)).booleanValue())
          {
            manager().peer(peerID).addNeighbor(this);
            addNeighbor(manager().peer(peerID));
          }

          network().rpc( new PeerMsgDisconnect(peerID(), ntr.peerID()) );
          disconnects_++;
        }
      }


      for (int i = 0; i < 5 && neighbors() > NetworkManager.MAX_CONNECTIONS; i++)
      {
        neighbor = pickNeighborToRemove();
        if (neighbor != null)
        {
          network().rpc( new PeerMsgDisconnect(peerID(), neighbor.peerID()) );
          disconnects_++;
        }
      }
    }
  }

  // peer leaves the network
  public void removeFromNetwork()
  {
    Vector  neighbors = (Vector) neighbors_.elementAt(cycle());

    // deregister with all neighbors
    for (int i = 0; i < neighbors(); i++)
      neighbor(i).removeNeighbor( this );
    neighbors.clear();
  }

}
