/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer.impl;

import java.util.*;
import java.io.*;

import qcsim.*;
import qcsim.peer.*;
import qcsim.peer.content.*;
import qcsim.net.*;

 /*
  * Title: Query Cycle Simulator
  * Description: P2P file sharing network simulator
  * Copyright: Copyright (c) 2002
  * Company:
  */

 /**
  * Creates and manage all peers in a simulation.
  *
  * @author unascribed
  * @version 1.0
  */

public class PeerManagerImpl implements PeerManager, Serializable
{

  /**
   * The query cycle manager.  Used to get other managers and
   * system parameters.
   */
  private QueryCycleManager manager_;

  /**
   * The content manager 
   */
  private PeerContentManager contentManager_;

  /**
   * list of peers in the network
   */
  private Hashtable peers_;

  /**
   * number of highly trusted nodes present in the network
   */
  private int highlyTrustedPeers_;

  /**
   * number of malicious present in the network
   */
  private int maliciousPeers_; 

  /**
   * number of good/normal nodes present in the network
   */
  private int goodPeers_;

  /**
   * Adaptive topology mode
   */
  private int adaptiveTopology_;

  private Vector clusterCoefficients_;

  private Vector diameters_;

  /**
   * Constructor
   * @param simulator Needed to obtain various simulation attributes.
   */
  public PeerManagerImpl(QueryCycleManager manager) 
  {
    manager_            = manager;
    contentManager_     = null;
    adaptiveTopology_   = AT_NONE;
    peers_              = new Hashtable();
    highlyTrustedPeers_ = 0;
    maliciousPeers_     = 0;
    goodPeers_          = 0;

    clusterCoefficients_ = new Vector();
    diameters_           = new Vector();

  }

    public void outputStats() {}

  public void adaptiveTopology(int mode)
  {
    adaptiveTopology_ = mode;
  }

  public int adaptiveTopology()
  {
    return adaptiveTopology_;
  }

  public Peer randomPeer(Random rand)
  {
    int id = rand.nextInt(peers());
    return peer(id); 
  }

  public NetworkManager network()
  {
    return manager_.networkManager();
  }

  public QueryCycleManager manager()
  {
    return manager_;
  }

  /**
   * Set the content manager.
   * @param the system content manager.
   */
  public void contentManager(PeerContentManager manager) 
  {
    contentManager_ = manager;

    for (int i = 0; i < peers(); i++)
    {
      if (peer(i).behavior().type() < PeerBehavior.MALICIOUS_PEER)
      {
        int pref = peer(i).behavior().categoryPreference();
        ((PeerImpl)peer(i)).content(manager.peerContent(pref));
      }
    }
    manager.distributeFiles();
  }

  public PeerContentManager contentManager()
  {
    return contentManager_;
  }

  /**
   * Number of peers in the system.
   * @return number of peers.
   */
  public int peers()
  {
    return peers_.size();
  }

  /**
   * Get peer
   * @param peerID the peer id.
   * @return the peer.
   */
  public Peer peer(int peerID)
  {
    return (Peer) peers_.get( new Integer(peerID) );
  }

    public Peer createPeer(int peerID, int behavior) {
	return new PeerImpl(this, peerID, behavior);
    }

  /**
   * Create a new peer
   * @param behavior the peer's behavior/type
   * @return the peer.
   */
  public Peer newPeer(int behavior) 
  {
    int      peerID = peers();
    Peer peer   = createPeer(peerID, behavior);

    switch (behavior)
    {
      case PeerBehavior.HIGHLY_TRUSTED_PEER:
        highlyTrustedPeers_++; break;
      case PeerBehavior.MALICIOUS_PEER:
        maliciousPeers_++; break;
      case PeerBehavior.GOOD_PEER:
        goodPeers_++; break;
    }

    peers_.put(new Integer(peerID), peer);

    return peer;
  }

  /**
   * Removes peer from the manager.
   * @param peerID the peer id.
   */
  public void deletePeer(int peerID)
  {
    peers_.remove(new Integer(peerID));
  }

  /**
   * Get the number of highly trusted peers in the system.
   * @return number of highly trusted peers.
   */
  public int highlyTrustedPeers()
  {
    return highlyTrustedPeers_;
  }

  /**
   * Get the number of malicious peers in the system.
   * @return number of malicious peers.
   */
  public int maliciousPeers()
  {
    return maliciousPeers_;
  }

  /**
   * Get the number of good peers in the system.
   * @return number of good peers.
   */
  public int goodPeers()
  {
    return goodPeers_;
  }

  public void freeRiders(double fraction)
  {
    if (fraction <= 0.0) return;

    for (int i = 0; i < goodPeers(); i++)
      peer(i).behavior().type(PeerBehavior.GOOD_PEER); 

    double inc = 1.0 / fraction;
    for (double i = 0; i < goodPeers(); i += inc)
    {
      int id = (int) Math.ceil( i );
      peer(id).behavior().type(PeerBehavior.FREERIDER_PEER); 
    }
  }

  public double characteristicPathLength(int peerType)
  {
    double charPathLength   = 0;
    double totalPathLengths = 0;
    double numberOfPaths    = 0;
    int diameter            = 5*diameter();

    for (int i = 0; i < peers(); i++)
    {
      ArrayList pathLengths = peer(i).shortestPaths();

      if (peerType > 0)
      {
        if (peerType < PeerBehavior.MALICIOUS_PEER && 
            peer(i).behavior().type() >= PeerBehavior.MALICIOUS_PEER)
          continue;  // goto next peer
        else if (peerType >= PeerBehavior.MALICIOUS_PEER && 
            peer(i).behavior().type() < PeerBehavior.MALICIOUS_PEER)
          continue;  // goto next peer
      }
 
      for (int j = 0; j < peers(); j++)
      {
        if (peer(j).behavior().type() >= PeerBehavior.MALICIOUS_PEER)
          break;  // I don't care about the path malicious -> *
        else if (i == j) 
          continue;

        int length = ((Integer)pathLengths.get(j)).intValue();

        if (length == Peer.INF)
          length = diameter;

        if (length > 0)
        {
          totalPathLengths += (double)length;
          numberOfPaths++;
        }
      }
    }

    if (numberOfPaths == 0) 
      return 0;
    charPathLength = totalPathLengths / numberOfPaths;

/*
    if (peerType == PeerBehavior.GOOD_PEER)
      System.err.println("GOOD PEER CHAR PATH: "+charPathLength);
    else if (peerType == PeerBehavior.MALICIOUS_PEER)
      System.err.println("MALICIOUS PEER CHAR PATH: "+charPathLength);
*/

    return (totalPathLengths / numberOfPaths);
  }

  public double avgClusterCoefficient()
  {
    double avgCluster     = 0;
    double sumCluster     = 0;

    while (clusterCoefficients_.size() <= manager_.cycle())
      clusterCoefficients_.addElement(new Double(-1));

    Double cc = (Double)clusterCoefficients_.elementAt(manager_.cycle());
    if (cc.doubleValue() >= 0.0)
      return cc.doubleValue();


    for (int i = 0; i < peers(); i++)
    {
      if (peer(i).behavior().type() < PeerBehavior.MALICIOUS_PEER)
      {
        double cluster = peer(i).clusterCoefficient();
        if (cluster > 0)
        {
          avgCluster += cluster;
          sumCluster++;
        }
      }
    }
    if (sumCluster > 0)
      avgCluster /= sumCluster;
    avgCluster /= 0.75;

    cc = new Double(avgCluster);
    clusterCoefficients_.setElementAt(cc, manager_.cycle());
    return cc.doubleValue();
  }

  public int diameter()
  {
    int diameter = 0;

    while (diameters_.size() <= manager_.cycle())
      diameters_.addElement(new Integer(-1));

    Integer d = (Integer)diameters_.elementAt(manager_.cycle());
    if (d.doubleValue() >= 0.0)
      return d.intValue();

    for (int i = 0; i < peers(); i++)
    {
      ArrayList paths = peer(i).shortestPaths();
      for (int j = 0; j < peer(i).peerID(); j++)
      {
        Integer length = (Integer) paths.get(j);
        if (length.intValue() > 0 && length.intValue() != Peer.INF)
        {
          diameter = (diameter < length.intValue()) ?
                      length.intValue() : diameter;
        }
      }
    }
    d = new Integer(diameter);
    diameters_.setElementAt(d, manager_.cycle());
    return d.intValue();
  }

  public int goodDownloads()
  {
    int numGoodDownloads = 0;

    for (int i = 0; i < peers(); i++)
    {
      if (peer(i).behavior().type() < PeerBehavior.MALICIOUS_PEER)
        numGoodDownloads += peer(i).goodDownloads();
    } 
    return numGoodDownloads;
  }

  public int badDownloads()
  {
    int numBadDownloads = 0;

    for (int i = 0; i < peers(); i++)
    {
      if (peer(i).behavior().type() < PeerBehavior.MALICIOUS_PEER)
        numBadDownloads += peer(i).inauthenticDownloads();
    } 
    return numBadDownloads;
  }

    public int voidDownloads()
    {
	int numVoidDownloads = 0;
	
	for (int i = 0; i < peers(); i++) {
	    if (peer(i).behavior().type() < PeerBehavior.MALICIOUS_PEER)
		numVoidDownloads += peer(i).voidDownloads();
	} 
	return numVoidDownloads;
    }

    public double totalScore()
    {
	double theTotalScore = 0;
	
	for (int i = 0; i < peers(); i++) {
	    if (peer(i).behavior().type() < PeerBehavior.MALICIOUS_PEER)
		theTotalScore += peer(i).totalScore();
	} 
	return theTotalScore;
    }

    public double totalExact()
    {
	double numExactMatch = 0;
	
	for (int i = 0; i < peers(); i++) {
	    if (peer(i).behavior().type() < PeerBehavior.MALICIOUS_PEER)
		numExactMatch += peer(i).exactMatches();
	} 
	return numExactMatch;
    }

  public double avgGoodUploads()
  {
    double numGoodUploads     = 0.0;
    double numPeersGoodUploads = 0.0;

    for (int i = 0; i < peers(); i++)
    {
      if (peer(i).behavior().type() < PeerBehavior.MALICIOUS_PEER &&
          peer(i).goodUploads() > 0)
      {
        numGoodUploads += peer(i).goodUploads();
        numPeersGoodUploads++;
      }
    } 
    return (numPeersGoodUploads == 0) ? 0 : numGoodUploads/numPeersGoodUploads; 
  }

  public int connectionTypes(int type)
  {
    int connections = 0;

    for (int i = 0; i < peers(); i++)
    {
      if (peer(i).behavior().type() <= PeerBehavior.GOOD_PEER ||
          peer(i).behavior().type() == PeerBehavior.HIGHLY_TRUSTED_PEER)
      {
        for (int n = 0; n < peer(i).neighbors(); n++)
        {
          if (type == PeerBehavior.MALICIOUS_PEER && 
              peer(i).neighbor(n).behavior().type() >= type)
            connections++;
          else if (peer(i).neighbor(n).behavior().type() == type)
            connections++;
        }
      }
    }
    return connections;
  }

  public double goodUploadRatio(int type)
  {
    double goodUploads  = 0; 
    double totalUploads = 0;

    for (int i = 0; i < peers(); i++)
    {
      if ((type == PeerBehavior.MALICIOUS_PEER && 
           peer(i).behavior().type() >= type) 
          ||
          peer(i).behavior().type() == type)
      {
        goodUploads  += peer(i).goodUploads();
        totalUploads += peer(i).goodUploads();
        totalUploads += peer(i).inauthenticUploads();
      }
    }
/*
    System.err.print("GOOD UPLOADS: "+ goodUploads);
    System.err.println("\tTOTAL UPLOADS: "+ totalUploads);
*/
    return (totalUploads == 0) ? 0 : goodUploads/totalUploads;
  }

// ======================= Create Collectives =============================

  /**
   * Create a collective.
   * @param peers Vector of peerIDs who will be in the collective
   * @param ctype The type of collective to create.
   * @param type The type of trust value configuration.
   */
  public void collective(Vector peers, int ctype, int type)
  {
    Vector collective = new Vector();
    Vector malicious  = new Vector();

    for (int i = 0; i < peers(); i++) 
    {
      Peer peer = peer(i);
      if (peers.contains(new Integer(peer.peerID())) &&
          peer.behavior().type() >= PeerBehavior.MALICIOUS_PEER)
      {
        collective.addElement(peer);
        peer.behavior().type(ctype);
      }
      else if (peer.behavior().type() >= PeerBehavior.MALICIOUS_PEER)
        malicious.addElement(peer);
    }

    setCollectiveTrust(collective, type);

    if (ctype == PeerBehavior.MALICIOUS_HIDDEN_SPY)
    {
      Vector tmp = collective;
      collective = malicious;
      malicious  = tmp;
      ctype      = PeerBehavior.MALICIOUS_SPY;

      for (int i = 0; i < malicious.size(); i++)
      {
        PeerImpl hidden = (PeerImpl) malicious.elementAt(i);
        hidden.removeFromNetwork(); 
        for (int j = 0; j < collective.size(); j++)
        {
          Peer spy = (Peer) collective.elementAt(j);
          spy.behavior().type(ctype);
          hidden.addNeighbor( spy );
          spy.addNeighbor( hidden );
        }
      }
    }

    if (ctype == PeerBehavior.MALICIOUS_CAMOUFLAGED ||
        ctype == PeerBehavior.MALICIOUS_SPY)
    {
      for (int i = 0; i < collective.size(); i++)
      {
        Peer p1 = (Peer) collective.elementAt(i);
        for (int j = 0; j < malicious.size(); j++)
        {
          Peer p2 = (Peer) malicious.elementAt(j);
          p1.behavior().collectiveNeighbor(p2.peerID());
          p1.localTrust(p2.peerID(), 1.0);
        }
      }
    }


  }

  /**
   * This will setup the collectives trust value configuration to
   * be one of MC_CHAIN, MC_CLIQUE, or MC_RANDOM
   * @param peers The list of peers in the collective.
   * @param type The trust value configuration type.
   */
  private void setCollectiveTrust(Vector peers, int type)
  {
    Peer p1 = null;
    Peer p2 = null;

    switch (type) 
    {
      case(MC_CHAIN): 
      {
        // malicious chain
        for (int i = 0; i < peers.size() - 1; i++) 
        {
          // update trust values
          p1 = (Peer)peers.elementAt(i);
          p2 = (Peer)peers.elementAt(i+1);
          p1.behavior().collectiveNeighbor( p2.peerID() );
          p1.localTrust(p2.peerID(), 1.0);
        }
        // have the last peer in the chain pointing to the first peer again
        p1 = (Peer)peers.lastElement();
        p2 = (Peer)peers.firstElement();
        p1.behavior().collectiveNeighbor( p2.peerID() );
        p1.localTrust(p2.peerID(), 1.0);
        break;
      }
      case(MC_CLIQUE): 
      {
        // malicious clique
        for (int i = 0; i < peers.size(); i++) 
        {
          p1 = (Peer)peers.elementAt(i);
          for (int j = 0; j < peers.size(); j++) 
          {
            if (i == j) continue;
            p2 = (Peer)peers.elementAt(j);
            p1.behavior().collectiveNeighbor( p2.peerID() );
            p1.localTrust(p2.peerID(), 1.0);
          }
        }
        break;
      }
      case(MC_RANDOM): 
      {
        // malicious chain
        int numberOfOpinions = (int)(0.3 * peers.size());
        for (int i = 0; i < peers.size(); i++) 
        {
          p1 = (Peer)peers.elementAt(i);
          for (int j = 0; j < numberOfOpinions; j++) 
          {
            int index;
            while (true) 
            {
              index = (int)(Math.random() * peers.size());
              p2 = (Peer)peers.elementAt(index);

              Vector cn = p1.behavior().collectiveNeighbors();
              if (!cn.contains(new Integer(p2.peerID())))
                break;
            }
            // update trust values. 
            p1.behavior().collectiveNeighbor( p2.peerID() );
            p1.localTrust(p2.peerID(), 1.0);
          }
        }
        break;
      }
      default: {
        System.out.println("Malicious collective type unknown.");
      }
    }
  }

//=====================================================================


  public void execCycle()
  {
    for (int i = 0; i < peers(); i++)
      ((PeerImpl)peer(i)).newCycle();

    for (int i = 0; i < peers(); i++)
      ((PeerImpl)peer(i)).query();
     
    for (int i = 0; i < 25; i++)
    {
      for (int j = 0; j < peers(); j++)
        ((PeerImpl)peer(j)).preSubmitTrust(i);
      for (int j = 0; j < peers(); j++)
        ((PeerImpl)peer(j)).submitTrust();
      for (int j = 0; j < peers(); j++)
        ((PeerImpl)peer(j)).postSubmitTrust();
    }
  }

}
