/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcsim.peer.impl;

import qcsim.peer.PeerManager;

public class RevenueTracker
{
    // Number of cycles you want to keep track over
    private static final int WINDOW_SIZE = 5;

    // payments made to this neighbor as an RTR seller
    // That is, payments made to this neighbor to buy RTRs
    private double sellRTRPayment[];
    private double totalSellRTRPayment;

    // revenue made from this neighbor as an RTR seller
    // That is, the resell and upload income from RTRs we bought
    // from this neighbor
    private double sellRTRRevenue[];
    private double totalSellRTRRevenue;

    // expected revenue made from this neigbhor as an RTR buyer
    private double buyRTRExpected[];
    private double totalBuyRTRExpected;

    // actual revenue made from this neighbor as an RTR buyer
    private double buyRTRRevenue[];
    private double totalBuyRTRRevenue;

    // how long has this peer been with us?
    private int trackTime;

    public RevenueTracker() {
	sellRTRPayment = new double[WINDOW_SIZE];
	sellRTRRevenue = new double[WINDOW_SIZE];
	buyRTRExpected = new double[WINDOW_SIZE];
	buyRTRRevenue = new double[WINDOW_SIZE];

	for (int i=0; i < WINDOW_SIZE; i++) {
	    sellRTRPayment[i] = 0;
	    sellRTRRevenue[i] = 0;
	    buyRTRExpected[i] = 0;
	    buyRTRRevenue[i] = 0;
	}
	totalSellRTRPayment = 0;
	totalSellRTRRevenue = 0;
	totalBuyRTRExpected = 0;
	totalBuyRTRRevenue = 0;

	trackTime = 0;
    }

    public void newCycle() {
	// shift everything over by one cycle
	totalSellRTRPayment -= sellRTRPayment[WINDOW_SIZE-1];
	totalSellRTRRevenue -= sellRTRRevenue[WINDOW_SIZE-1];
	totalBuyRTRExpected -= buyRTRExpected[WINDOW_SIZE-1];
	totalBuyRTRRevenue -= buyRTRRevenue[WINDOW_SIZE-1];

	for (int i=0; i < WINDOW_SIZE-1; i++) {
	    sellRTRPayment[i+1] = sellRTRPayment[i];
	    sellRTRRevenue[i+1] = sellRTRRevenue[i];
	    buyRTRExpected[i+1] = buyRTRExpected[i];
	    buyRTRRevenue[i+1] = buyRTRRevenue[i];
	}
	trackTime++;
    }

    public void addSellRTRPayment(double p) {
	sellRTRPayment[0] += p;
	totalSellRTRPayment += p;
    }

    public void addSellRTRRevenue(double p) {
	sellRTRRevenue[0] += p;
	totalSellRTRRevenue += p;
    }

    public void addBuyRTRExpected(double p) {
	buyRTRExpected[0] += p;
	totalBuyRTRExpected += p;
    }

    public void addBuyRTRRevenue(double p) {
	buyRTRRevenue[0] += p;
	totalBuyRTRRevenue += p;
    }

    public double getSellRTRPayment() { return totalSellRTRPayment;}
    public double getSellRTRRevenue() { return totalSellRTRRevenue;}
    public double getBuyRTRExpected() { return totalBuyRTRExpected;}
    public double getBuyRTRRevenue() { return totalBuyRTRRevenue;}

    public double getProfit(int profitType) {
	if (trackTime == 0)
	    return 0;

	double profit = 0;
	switch (profitType) {
	case PeerManagerRTR.PROFIT_SELLER:
	    profit = (totalSellRTRRevenue - totalSellRTRPayment) /
		(double) trackTime;
	    break;
	case PeerManagerRTR.PROFIT_BUYER:
	    profit = (totalBuyRTRRevenue - totalBuyRTRExpected) /
		(double) trackTime;
	    break;
	case PeerManagerRTR.PROFIT_BOTH:
	    profit = (totalBuyRTRRevenue + totalSellRTRRevenue -
		       totalSellRTRPayment - totalBuyRTRExpected) /
		(double) trackTime;
	    // should we include the total expected revenue in this equation?
	    break;
	}
	return profit;
    }
}
