/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcui.view;

import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.plaf.basic.*;

import qcui.util.*;

import qcsim.*;
import qcsim.peer.*;
import qcsim.util.*;

public class LinkGraphView extends QueryCycleView implements ActionListener
{
  private JCheckBox colorTrust_;

  private double[] avgLinkProbabilityContent_;

  public LinkGraphView(QueryCycleManager manager, String resource)
  {
    super(manager);

    avgLinkProbabilityContent_ = new double[10];
    for (int i = 0; i < 10; i++)
      avgLinkProbabilityContent_[i] = 0.0;

    colorTrust_   = new JCheckBox("Color Trust");
    colorTrust_.addActionListener( this );

    JPanel checkBoxPanel = new JPanel();
    checkBoxPanel.setLayout(new BasicOptionPaneUI.ButtonAreaLayout(true, 0));
    checkBoxPanel.add(colorTrust_);

    controlPanel().add(checkBoxPanel);

  }

  public void paintComponent(Graphics g)
  {
    super.paintComponent( g );
    super.clearBackground( g );

    controlPanel().setBorder(BorderFactory.createTitledBorder(viewName()));
   
    Dimension dim  = getSize();
    int height = (int) (dim.getHeight() * 0.4);
    int width  = (int) (dim.getWidth() * 0.4);

    int graphY = (int)(dim.getHeight() * 0.45); 
    int graphX = (int)(dim.getWidth() * 0.10);

    Plot plot = manager().plot(QueryCycleManager.LINK_PROBABILITY);

    plot.graph(g, this, Plot.LINE_GRAPH, 10, 10, 
               graphX, graphY, new Dimension(width, height));

    // Links vs. Local Trust

    width  = (int) (dim.getWidth() * 0.4);
    graphX = (int)(dim.getWidth() * 0.55);

    plot = manager().plot(QueryCycleManager.LINK_TRUST);

    plot.graph(g, this, Plot.LINE_GRAPH, 10, 10, 
               graphX, graphY, new Dimension(width, height));


    // Draw Link vs. Data Volume.
    height = (int) (dim.getHeight() * 0.35);
    width  = (int) (dim.getWidth() * 0.8);

    graphY = (int)(dim.getHeight() * 0.90); 
    graphX = (int)(dim.getWidth() * 0.10);

    int xpos=0, ypos=0;

    g.setColor(Color.black);
    g.drawLine(graphX,graphY,graphX+width,graphY);   // Draw x axis
    g.drawLine(graphX,graphY,graphX,graphY-height);  // Draw y axis

    int maxUpload = 0;

    int maxLinks = 0;
    int temp     = 0;
    for (int i = 0; i < peers(); i++)
    {
      if (peer(i).behavior().type() >= PeerBehavior.MALICIOUS_PEER)
        break;

      temp = peer(i).totalGoodUploads();
      if (maxUpload < temp) maxUpload = temp;
      temp = peer(i).neighbors();
      if (maxLinks < temp) maxLinks = temp;
    }
    maxUpload += 10;
    
    int lstep = (int) Math.ceil(maxLinks/10.0);
    for (int i = 1; i <= 10; i++)
    {
      xpos = graphX + ((i * width) / 10);
      ypos = graphY - ((i * height) / 10);

      g.drawString(Integer.toString((int)((i*maxUpload)/10)), 
                   xpos, graphY+15);

      g.drawString(Integer.toString(i * lstep), graphX-15, ypos);

    }
    VTextIcon axis = new VTextIcon(new JPanel(),"Connections",
                                   VTextIcon.ROTATE_LEFT);
    axis.paintIcon(this, g, graphX-30, graphY-(3*height/4));
    g.drawString("Number of Authentic Uploads", graphX+(width/2), graphY+30);

    for (int i = 0; i < peers(); i++)
    {
      Peer peer = peer(i);

      if (peer.behavior().type() >= PeerBehavior.MALICIOUS_PEER)
        break;

      xpos = graphX + ((width * peer.totalGoodUploads())/maxUpload);
      ypos = graphY - ((height * peer.neighbors()) / (10 * lstep));

      g.setColor(peer.description().color(colorTrust_.isSelected()));
      g.fillRect(xpos+2, ypos-2, 4, 4);
    }

    super.notifyDone();
  } 

  public String viewName() 
  { 
    return "Link Graph View";
  }

  public void actionPerformed(ActionEvent event) 
  {
    validate();
    repaint();
  }

  private double l1Norm(double [] d1, double [] d2)
  {
    if (d1.length != d2.length)
    {
      System.err.println("L1 Norm error: length "+d1.length+" != "+d2.length);
      System.exit(1);
    }

    double norm = 0.0;
    for (int i = 0; i < d1.length; i++)
    {
      norm += (d1[i] <= d2[i]) ? d2[i] - d1[i] : d1[i] - d2[i];
    }
    if (norm > 2.0) norm = 2.0;
    return norm;
  }

}
