/* ====================================================================
 * The Stanford P2P Sociology Project Software License, Version 1.0
 *
 * Copyright (c) 2003 The Stanford P2P Sociology Project.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Stanford P2P Sociology Project(http://p2p.stanford.edu/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * For more
 * information on the Stanford P2P Sociology Project,
 * see <http://p2p.stanford.edu/>.
 *
 */
package qcui.view;

import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.plaf.basic.*;

import qcui.util.*;

import qcsim.*;
import qcsim.peer.*;
import qcsim.peer.content.*;
import qcsim.util.*;


public class TopologyView extends QueryCycleView 
       implements ActionListener, MouseListener
{
  private static final String TOPOLOGY_PATH_LENGTH = "Avg. Path Length";
  private static final String TOPOLOGY_DIAMETER    = "Network Diameter";
  private static final String TOPOLOGY_CONNECTIONS = "Network Connections";

  private JLabel avgPathLength_;

  private JLabel networkDiameter_;

  private JLabel connections_;

  protected Peer selectedPeer_;

  private Color[] contentPieColors_;

  public TopologyView(QueryCycleManager manager, boolean describe)
  {
    super(manager);
    addMouseListener( this );
    selectedPeer_ = null;

    if (describe)
    {
      Dimension dim  = new Dimension(getWidth(), getHeight());
      for (int i = 0; i < peers(); i++)
        peer(i).description().displayDescription(dim, Math.random());
    }

    controlPanel().setLayout(new BoxLayout(controlPanel(), BoxLayout.Y_AXIS));

    JPanel labels = new JPanel();
    labels.setLayout(new GridLayout(5,1));

    avgPathLength_ = new JLabel();
    labels.add(avgPathLength_);

    networkDiameter_ = new JLabel();
    labels.add(networkDiameter_);

    connections_ = new JLabel();
    labels.add(connections_);

    controlPanel().add(labels);
  }

  public TopologyView(QueryCycleManager manager)
  {
    this(manager, false);

    JPanel select = new JPanel();
    select.setLayout(new BasicOptionPaneUI.ButtonAreaLayout(true, 0));

    JComboBox combo = new JComboBox();
    combo.addItem("Adaptive Topology Off");
    combo.addItem("AT Local Trust");
    combo.addItem("AT Connection Trust");
    combo.addItem("AT All");
    combo.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent event) 
      {
        JComboBox cb = (JComboBox) event.getSource();
        manager().peerManager().adaptiveTopology(cb.getSelectedIndex() );
      }
    });
    select.add(combo);

    controlPanel().add(new JLabel(""));
    controlPanel().add(select);
  }

  public void paintComponent(Graphics g)
  {
    super.paintComponent( g );
    super.drawBackground(g);
    controlPanel().setBorder(BorderFactory.createTitledBorder(viewName()));
   
    if (selectedPeer_ != null)
    {
      drawPeerInfo( g, selectedPeer_);
    }
    else
    {
      ArrayList paths   = null;
      int connections   = 0;
      int    diameter   = 0;

      double sumPathLengths = 0;
      double numPaths       = 0;

      for (int i = 0; i < peers(); i++)
      {
        connections += peer(i).neighbors();
        paths        = peer(i).shortestPaths();
        for (int j = 0; j < peer(i).peerID(); j++)
        {
          Integer length = (Integer) paths.get(j);
          if (length.intValue() > 0 && length.intValue() != Peer.INF)
          {
            numPaths++;
            sumPathLengths += (double)length.intValue();
            diameter = (diameter < length.intValue()) ?
                        length.intValue() : diameter;
          }
        }

        peer(i).description().drawLinks(g);
      }

      for (int i = 0; i < peers(); i++)
        peer(i).description().draw(g, peer(i).description().color(false));

      String avgPath = Double.toString(sumPathLengths / numPaths);
      if (avgPath.length() > 4)
         avgPath = avgPath.substring(0, 5);
      avgPathLength_.setText("Avg. Path Length: "+avgPath); 

      networkDiameter_.setText("Network Diameter: "+diameter);
      connections_.setText("Total Connections: "+connections);

    }

    notifyDone();
  } 

  public String viewName() 
  { 
    return (selectedPeer_ == null) ? 
           "Topology View: All Peers" :
           "Topology View: Peer "+selectedPeer_.peerID(); 
  }

  public void actionPerformed(ActionEvent event) 
  {
    validate();
    repaint();
  }

  private void drawPeerInfo(Graphics g, Peer peer)
  {
    if (peer.behavior().type() >= PeerBehavior.MALICIOUS_PEER)
      return;
    else if (peer.content() == null)
      return;
    
    PeerContent peerContent = peer.content();
    VTextIcon   id;
    int         scaler    = 1;
    int         xpos      = 0;
    int         ypos      = 0;
    PieValue[] slices       = new PieValue[peerContent.categories()];
    PieValue[] sortedSlices = new PieValue[peerContent.categories()];

    Dimension dim = getSize();

    int height = (int) (dim.getHeight() * 0.50);
    int width  = (int) (dim.getWidth() * 0.85);

    int graphY = (int)(dim.getHeight() * 0.90); 
    int graphX = (int)(dim.getWidth() * 0.10);

    g.setColor(Color.white);
    g.fillRect(0,0,getWidth(),getHeight());

    g.setColor(Color.black);
    g.drawLine(graphX,graphY,graphX+width+10,graphY);       // Draw x axis
    g.drawLine(graphX,graphY,graphX,graphY-height);  // Draw y axis

    double maxVolume = 0.0;
    for (int index = 0; index < peerContent.categories(); index++)
    {
      Content category = peerContent.category(index);
      maxVolume = (double) ((category.volume() < maxVolume) ? 
                            maxVolume : category.volume());

      slices[index] = new PieValue(category.volume(), category.category());
    }
    scaler = (int) Math.ceil(maxVolume / height);

    for (int i = 0; i < slices.length; i++) 
    {
      int min  = Integer.MAX_VALUE;
      int next = 0;
      for (int j = 0; j < slices.length; j++)
      {
        if (slices[j] != null && slices[j].category < min)
        {
          min  = slices[j].category;
          next = j; 
        }
      }
      sortedSlices[i] = slices[next];
      slices[next]    = null;
    }

    drawPie(g, new Rectangle(10,10,height/2,height/2), sortedSlices);
    

    PeerContentManager cm = peerContent.manager();
    xpos = (3*width) / 4;
    ypos = 15;
    g.drawString("Content Category", xpos, ypos);
    g.drawString("Unique Files", xpos+110, ypos);
    g.drawLine(xpos, ypos+3, xpos+200, ypos+3); 
    g.drawLine(xpos+100, 5, xpos+100, height);

    for (int i = 1; i <= cm.contentCategories(); i++)
    {
      ypos += 15;
      int uf = cm.uniqueFiles(i);
      g.drawString(Integer.toString(i), xpos, ypos);
      g.drawString(Integer.toString(uf), xpos+110, ypos);
    }

    // Draw Content Graph.
    for (int i = 1; i <= 10; i++)
    {
      ypos = graphY - ((i*height) / 10);
      g.drawLine(graphX-5,ypos,graphX+5,ypos);       // Draw x axis
      g.drawString(Integer.toString(i*height*scaler/10), graphX-25, ypos);
    }

    for (int i = 1; i <= cm.contentCategories(); i++)
    {
      Content category = peerContent.findCategory(i);
      xpos = (int)((width*i)/cm.contentCategories()) + graphX;
      id = new VTextIcon(new JPanel(),Integer.toString(i),VTextIcon.ROTATE_LEFT);
      id.paintIcon(this, g, xpos-4, graphY+10);

      g.drawLine(xpos+4,graphY,xpos+4,graphY+5);

      if (category != null)
      {
        int bar  = category.volume() / scaler;
        ypos = graphY - bar;

        g.fillRect(xpos, ypos, 8, bar);
      }
    }

    id = new VTextIcon(new JPanel(), "Number of Files", VTextIcon.ROTATE_LEFT);
    id.paintIcon(this, g, 20, graphY-(3*height/4));

    g.drawString("Content Category", graphX+(width/2), graphY+50);

  }

  // slices is an array of values that represent the size of each slice.
  private void drawPie(Graphics g, Rectangle area, PieValue[] slices) 
  {
    if (contentPieColors_ == null) initContentPieColors();

    // Get total value of all slices
    double total = 0.0D;
    for (int i=0; i < slices.length; i++) 
      total += slices[i].value;
    
    // Draw each pie slice
    Color  c;
    double curValue   = 0.0D;
    int    startAngle = 0;
    for (int i = 0; i < slices.length; i++) 
    {
      // Compute the start and stop angles
      startAngle   = (int)(curValue * 360 / total);
      int arcAngle = (int)(slices[i].value * 360 / total);
    
      /* Ensure that rounding errors do not leave a gap between 
       * the first and last slice
       */
      if (i == slices.length-1) {
        arcAngle = 360 - startAngle;
      }
    
      // Set the color and draw a filled arc
      g.setColor(contentPieColors_[slices[i].category-1]);
      g.fillArc(area.x, area.y, area.width, area.height, startAngle, arcAngle);

      g.fillRect(area.x+area.width+5, area.y+(i*20), 10, 10);
      g.setColor(Color.black);
      g.drawString("Content Category "+slices[i].category+".",
                   area.x+area.width+18, area.y + (i * 20) + 10);
      g.drawString((int)(Math.ceil(100*(slices[i].value/total)))+"%", 
                   area.x+area.width+150, area.y + (i * 20) + 10);
    
      curValue += slices[i].value;
    }
  }

  public void initContentPieColors() 
  {
    Random rand = new Random();
    PeerContentManager pcm = peerManager().contentManager();

    contentPieColors_ = new Color[pcm.contentCategories()];

    for (int i = 0; i < contentPieColors_.length; i++)
    {
      contentPieColors_[i] = 
        new Color(rand.nextInt(255), rand.nextInt(255), rand.nextInt(255));
    }

  }

  public void mouseClicked(MouseEvent e) 
  {
    selectedPeer_ = null;

    for (int i = 0; i < peers(); i++)
    {
      Peer p = (Peer) peer(i);
      if (p.description().contains(e.getX(), e.getY()))
      {
        selectedPeer_ = p;
        
        break;
      }
    }
    repaint();
  }

  public void mouseEntered(MouseEvent e) {}

  public void mouseExited(MouseEvent e) {} 

  public void mousePressed(MouseEvent e) { selectedPeer_ = null; }

  public void mouseReleased(MouseEvent e) { }

  // Class to hold a value for a slice
  private class PieValue 
  {
    public double value;
    public int    category;
    
    public PieValue(double value, int category) 
    {
      this.value    = value;
      this.category = category;
    }
  }
}
